import numpy as np
import pandas as pd
import sklearn.metrics as skm2

try:
    import nnetsauce as ns
except:
    pass
try:
    import GPopt as gp
except:
    pass
from collections import namedtuple
from copy import deepcopy
from tqdm import tqdm
from sklearn import metrics
from sklearn.base import ClassifierMixin, RegressorMixin
try: 
    from sklearn.utils import all_estimators
except ImportError:
    pass
from sklearn.model_selection import cross_val_score
from ..custom import CustomClassifier
from ..predictionset import PredictionSet
from ..utils import matrixops as mo


class DeepClassifier(CustomClassifier, ClassifierMixin):
    """
    Deep Classifier

    Parameters:

        obj: an object
            A base learner, see also https://www.researchgate.net/publication/380701207_Deep_Quasi-Randomized_neural_Networks_for_classification

        n_layers: int (default=3)
            Number of layers. `n_layers = 1` is a simple `CustomClassifier`

        verbose : int, optional (default=0)
            Monitor progress when fitting.

        All the other parameters are nnetsauce `CustomClassifier`'s

    Examples:

        ```python
        import nnetsauce as ns
        from sklearn.datasets import load_breast_cancer
        from sklearn.model_selection import train_test_split
        from sklearn.linear_model import LogisticRegressionCV
        data = load_breast_cancer()
        X = data.data
        y= data.target
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=.2, random_state=123)
        obj = LogisticRegressionCV()
        clf = ns.DeepClassifier(obj)
        clf.fit(X_train, y_train)
        print(clf.score(clf.predict(X_test), y_test))
        ```
    """

    def __init__(
        self,
        obj,
        # Defining depth
        n_layers=3,
        verbose=0,
        # CustomClassifier attributes
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        level=None,
        pi_method="icp",
        seed=123,
        backend="cpu",
    ):
        super().__init__(
            obj=obj,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            level=level,
            pi_method=pi_method,
            seed=seed,
            backend=backend,
        )

        assert n_layers >= 1, "must have n_layers >= 1"

        self.stacked_obj = obj
        self.verbose = verbose
        self.n_layers = n_layers

    def fit(self, X, y, sample_weight=None, **kwargs):
        """Fit Classification algorithms to X and y.
        Parameters
        ----------
        X : array-like,
            Training vectors, where rows is the number of samples
            and columns is the number of features.
        y : array-like,
            Training vectors, where rows is the number of samples
            and columns is the number of features.
        sample_weight: array-like, shape = [n_samples]
                Sample weights.
        Returns
        -------
        A fitted object
        """

        self.classes_ =  np.unique(y)
        self.n_classes_ = len(self.classes_)  # for compatibility with         scikit-learn

        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X)

        # init layer
        self.stacked_obj = CustomClassifier(
            obj=self.stacked_obj,
            n_hidden_features=self.n_hidden_features,
            activation_name=self.activation_name,
            a=self.a,
            nodes_sim=self.nodes_sim,
            bias=self.bias,
            dropout=self.dropout,
            direct_link=self.direct_link,
            n_clusters=self.n_clusters,
            cluster_encode=self.cluster_encode,
            type_clust=self.type_clust,
            type_scaling=self.type_scaling,
            col_sample=self.col_sample,
            row_sample=self.row_sample,
            seed=self.seed,
            backend=self.backend,
        )

        if self.verbose > 0:
            iterator = tqdm(range(self.n_layers - 1))
        else:
            iterator = range(self.n_layers - 1)

        for _ in iterator:
            self.stacked_obj = deepcopy(
                CustomClassifier(
                    obj=self.stacked_obj,
                    n_hidden_features=self.n_hidden_features,
                    activation_name=self.activation_name,
                    a=self.a,
                    nodes_sim=self.nodes_sim,
                    bias=self.bias,
                    dropout=self.dropout,
                    direct_link=self.direct_link,
                    n_clusters=self.n_clusters,
                    cluster_encode=self.cluster_encode,
                    type_clust=self.type_clust,
                    type_scaling=self.type_scaling,
                    col_sample=self.col_sample,
                    row_sample=self.row_sample,
                    seed=self.seed,
                    backend=self.backend,
                )
            )

        if self.level is not None:
            self.stacked_obj = PredictionSet(
                obj=self.stacked_obj, method=self.pi_method, level=self.level
            )

        try:
            self.stacked_obj.fit(X, y, sample_weight=sample_weight, **kwargs)
        except Exception as e:
            self.stacked_obj.fit(X, y)

        self.obj = deepcopy(self.stacked_obj)

        return self.obj

    def predict(self, X):
        return self.obj.predict(X)

    def predict_proba(self, X):
        return self.obj.predict_proba(X)

    def score(self, X, y, scoring=None):
        return self.obj.score(X, y, scoring)

    def cross_val_optim(
        self,
        X_train,
        y_train,
        X_test=None,
        y_test=None,
        scoring="accuracy",
        surrogate_obj=None,
        cv=5,
        n_jobs=None,
        n_init=10,
        n_iter=190,
        abs_tol=1e-3,
        verbose=2,
        seed=123,
        **kwargs,
    ):
        """Cross-validation function and hyperparameters' search

        Parameters:

            X_train: array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            y_train: array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            X_test: array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

            y_test: array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

            scoring: str
                scoring metric; see https://scikit-learn.org/stable/modules/model_evaluation.html#the-scoring-parameter-defining-model-evaluation-rules

            surrogate_obj: an object;
                An ML model for estimating the uncertainty around the objective function

            cv: int;
                number of cross-validation folds

            n_jobs: int;
                number of jobs for parallel execution

            n_init: an integer;
                number of points in the initial setting, when `x_init` and `y_init` are not provided

            n_iter: an integer;
                number of iterations of the minimization algorithm

            abs_tol: a float;
                tolerance for convergence of the optimizer (early stopping based on acquisition function)

            verbose: int
                controls verbosity

            seed: int
                reproducibility seed

            **kwargs: dict
                additional parameters to be passed to the estimator

        Examples:

            ```python
            ```
        """

        num_to_activation_name = {1: "relu", 2: "sigmoid", 3: "tanh"}
        num_to_nodes_sim = {1: "sobol", 2: "uniform", 3: "hammersley"}
        num_to_type_clust = {1: "kmeans", 2: "gmm"}

        def deepclassifier_cv(
            X_train,
            y_train,
            # Defining depth
            n_layers=3,
            # CustomClassifier attributes
            n_hidden_features=5,
            activation_name="relu",
            nodes_sim="sobol",
            dropout=0,
            n_clusters=2,
            type_clust="kmeans",
            cv=5,
            n_jobs=None,
            scoring="accuracy",
            seed=123,
        ):
            self.set_params(
                **{
                    "n_layers": n_layers,
                    # CustomClassifier attributes
                    "n_hidden_features": n_hidden_features,
                    "activation_name": activation_name,
                    "nodes_sim": nodes_sim,
                    "dropout": dropout,
                    "n_clusters": n_clusters,
                    "type_clust": type_clust,
                    **kwargs,
                }
            )
            return -cross_val_score(
                estimator=self,
                X=X_train,
                y=y_train,
                scoring=scoring,
                cv=cv,
                n_jobs=n_jobs,
                verbose=0,
            ).mean()

        # objective function for hyperparams tuning
        def crossval_objective(xx):
            return deepclassifier_cv(
                X_train=X_train,
                y_train=y_train,
                # Defining depth
                n_layers=int(np.ceil(xx[0])),
                # CustomClassifier attributes
                n_hidden_features=int(np.ceil(xx[1])),
                activation_name=num_to_activation_name[np.ceil(xx[2])],
                nodes_sim=num_to_nodes_sim[int(np.ceil(xx[3]))],
                dropout=xx[4],
                n_clusters=int(np.ceil(xx[5])),
                type_clust=num_to_type_clust[int(np.ceil(xx[6]))],
                cv=cv,
                n_jobs=n_jobs,
                scoring=scoring,
                seed=seed,
            )

        if surrogate_obj is None:
            gp_opt = gp.GPOpt(
                objective_func=crossval_objective,
                lower_bound=np.array([0, 3, 0, 0, 0.0, 0, 0]),
                upper_bound=np.array([5, 100, 3, 3, 0.4, 5, 2]),
                params_names=[
                    "n_layers",
                    # CustomClassifier attributes
                    "n_hidden_features",
                    "activation_name",
                    "nodes_sim",
                    "dropout",
                    "n_clusters",
                    "type_clust",
                ],
                method="bayesian",
                n_init=n_init,
                n_iter=n_iter,
                seed=seed,
            )
        else:
            gp_opt = gp.GPOpt(
                objective_func=crossval_objective,
                lower_bound=np.array([0, 3, 0, 0, 0.0, 0, 0]),
                upper_bound=np.array([5, 100, 3, 3, 0.4, 5, 2]),
                params_names=[
                    "n_layers",
                    # CustomClassifier attributes
                    "n_hidden_features",
                    "activation_name",
                    "nodes_sim",
                    "dropout",
                    "n_clusters",
                    "type_clust",
                ],
                acquisition="ucb",
                method="splitconformal",
                surrogate_obj=ns.PredictionInterval(
                    obj=surrogate_obj, method="splitconformal"
                ),
                n_init=n_init,
                n_iter=n_iter,
                seed=seed,
            )

        res = gp_opt.optimize(verbose=verbose, abs_tol=abs_tol)
        res.best_params["n_layers"] = int(np.ceil(res.best_params["n_layers"]))
        res.best_params["n_hidden_features"] = int(
            np.ceil(res.best_params["n_hidden_features"])
        )
        res.best_params["activation_name"] = num_to_activation_name[
            np.ceil(res.best_params["activation_name"])
        ]
        res.best_params["nodes_sim"] = num_to_nodes_sim[
            int(np.ceil(res.best_params["nodes_sim"]))
        ]
        res.best_params["dropout"] = res.best_params["dropout"]
        res.best_params["n_clusters"] = int(
            np.ceil(res.best_params["n_clusters"])
        )
        res.best_params["type_clust"] = num_to_type_clust[
            int(np.ceil(res.best_params["type_clust"]))
        ]

        # out-of-sample error
        if X_test is not None and y_test is not None:
            self.set_params(**res.best_params, verbose=0, seed=seed)
            preds = self.fit(X_train, y_train).predict(X_test)
            # check error on y_test
            oos_err = getattr(metrics, scoring + "_score")(
                y_true=y_test, y_pred=preds
            )
            result = namedtuple("result", res._fields + ("test_" + scoring,))
            return result(*res, oos_err)
        else:
            return res

    def lazy_cross_val_optim(
        self,
        X_train,
        y_train,
        X_test=None,
        y_test=None,
        scoring="accuracy",
        surrogate_objs=None,
        customize=False,
        cv=5,
        n_jobs=None,
        n_init=10,
        n_iter=190,
        abs_tol=1e-3,
        verbose=1,
        seed=123,
    ):
        """Automated Cross-validation function and hyperparameters' search using multiple surrogates

        Parameters:

            X_train: array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            y_train: array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            X_test: array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

            y_test: array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

            scoring: str
                scoring metric; see https://scikit-learn.org/stable/modules/model_evaluation.html#the-scoring-parameter-defining-model-evaluation-rules

            surrogate_objs: object names as a list of strings;
                ML models for estimating the uncertainty around the objective function

            customize: boolean
                if True, the surrogate is transformed into a quasi-randomized network (default is False)

            cv: int;
                number of cross-validation folds

            n_jobs: int;
                number of jobs for parallel execution

            n_init: an integer;
                number of points in the initial setting, when `x_init` and `y_init` are not provided

            n_iter: an integer;
                number of iterations of the minimization algorithm

            abs_tol: a float;
                tolerance for convergence of the optimizer (early stopping based on acquisition function)

            verbose: int
                controls verbosity

            seed: int
                reproducibility seed

        Examples:

            ```python
            ```
        """

        removed_regressors = [
            "TheilSenRegressor",
            "ARDRegression",
            "CCA",
            "GaussianProcessRegressor",
            "GradientBoostingRegressor",
            "HistGradientBoostingRegressor",
            "IsotonicRegression",
            "MultiOutputRegressor",
            "MultiTaskElasticNet",
            "MultiTaskElasticNetCV",
            "MultiTaskLasso",
            "MultiTaskLassoCV",
            "OrthogonalMatchingPursuit",
            "OrthogonalMatchingPursuitCV",
            "PLSCanonical",
            "PLSRegression",
            "RadiusNeighborsRegressor",
            "RegressorChain",
            "StackingRegressor",
            "VotingRegressor",
        ]

        results = []

        for est in all_estimators():

            if surrogate_objs is None:

                if issubclass(est[1], RegressorMixin) and (
                    est[0] not in removed_regressors
                ):
                    try:
                        if customize == True:
                            print(f"\n surrogate: CustomClassifier({est[0]})")
                            surr_obj = ns.CustomClassifier(obj=est[1]())
                        else:
                            print(f"\n surrogate: {est[0]}")
                            surr_obj = est[1]()
                        res = self.cross_val_optim(
                            X_train=X_train,
                            y_train=y_train,
                            X_test=X_test,
                            y_test=y_test,
                            surrogate_obj=surr_obj,
                            cv=cv,
                            n_jobs=n_jobs,
                            scoring=scoring,
                            n_init=n_init,
                            n_iter=n_iter,
                            abs_tol=abs_tol,
                            verbose=verbose,
                            seed=seed,
                        )
                        print(f"\n result: {res}")
                        if customize == True:
                            results.append((f"CustomClassifier({est[0]})", res))
                        else:
                            results.append((est[0], res))
                    except:
                        pass

            else:

                if (
                    issubclass(est[1], RegressorMixin)
                    and (est[0] not in removed_regressors)
                    and est[0] in surrogate_objs
                ):
                    try:
                        if customize == True:
                            print(f"\n surrogate: CustomClassifier({est[0]})")
                            surr_obj = ns.CustomClassifier(obj=est[1]())
                        else:
                            print(f"\n surrogate: {est[0]}")
                            surr_obj = est[1]()
                        res = self.cross_val_optim(
                            X_train=X_train,
                            y_train=y_train,
                            X_test=X_test,
                            y_test=y_test,
                            surrogate_obj=surr_obj,
                            cv=cv,
                            n_jobs=n_jobs,
                            scoring=scoring,
                            n_init=n_init,
                            n_iter=n_iter,
                            abs_tol=abs_tol,
                            verbose=verbose,
                            seed=seed,
                        )
                        print(f"\n result: {res}")
                        if customize == True:
                            results.append((f"CustomClassifier({est[0]})", res))
                        else:
                            results.append((est[0], res))
                    except:
                        pass

        return results



import numpy as np
import pandas as pd
import time

try: 
    from sklearn.utils import all_estimators
except ImportError:
    pass
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import StandardScaler, OneHotEncoder, OrdinalEncoder
from sklearn.compose import ColumnTransformer
from sklearn.base import RegressorMixin
from sklearn.metrics import  mean_absolute_error
try: 
    from sklearn.metrics import mean_pinball_loss
except ImportError:
    pass

from tqdm import tqdm

from .config import DEEPREGRESSORSMTS, REGRESSORSMTS
from ..deep import DeepMTS
from ..mts import ClassicalMTS, MTS
from ..utils import (
    convert_df_to_numeric,
    coverage,
    dict_to_dataframe_series,
    mean_errors,
    winkler_score,
)

import warnings

warnings.filterwarnings("ignore")
pd.set_option("display.precision", 2)
pd.set_option("display.float_format", lambda x: "%.2f" % x)

numeric_transformer = Pipeline(
    steps=[
        ("imputer", SimpleImputer(strategy="mean")),
        ("scaler", StandardScaler()),
    ]
)

try:
    categorical_transformer_low = Pipeline(
        steps=[
            (
                "imputer",
                SimpleImputer(strategy="constant", fill_value="missing"),
            ),
            (
                "encoding",
                OneHotEncoder(handle_unknown="ignore", sparse_output=False),
            ),
        ]
    )
except TypeError:
    categorical_transformer_low = Pipeline(
        steps=[
            (
                "imputer",
                SimpleImputer(strategy="constant", fill_value="missing"),
            ),
            (
                "encoding",
                OneHotEncoder(handle_unknown="ignore", sparse=False),
            ),
        ]
    )

categorical_transformer_high = Pipeline(
    steps=[
        ("imputer", SimpleImputer(strategy="constant", fill_value="missing")),
        # 'OrdinalEncoder' Raise a ValueError when encounters an unknown value. Check https://github.com/scikit-learn/scikit-learn/pull/13423
        ("encoding", OrdinalEncoder()),
    ]
)


# Helper function


def get_card_split(df, cols, n=11):
    cond = df[cols].nunique() > n
    card_high = cols[cond]
    card_low = cols[~cond]
    return card_low, card_high


def adjusted_rsquared(r2, n, p):
    return 1 - (1 - r2) * ((n - 1) / (n - p - 1))


# def mean_percentage_error(y_pred, y_true):
#     pe = (y_pred - y_true) / np.maximum(np.abs(y_true),
#                                         np.finfo(np.float64).eps)
#     return np.average(pe)


class LazyDeepMTS(MTS):
    """

    Fitting -- almost -- all the regression algorithms with layers of
    nnetsauce's CustomRegressor to multivariate time series
    and returning their scores.

    Parameters:

        verbose: int, optional (default=0)
            Any positive number for verbosity.

        ignore_warnings: bool, optional (default=True)
            When set to True, the warning related to algorigms that are not
            able to run are ignored.

        custom_metric: function, optional (default=None)
            When function is provided, models are evaluated based on the custom
              evaluation metric provided.

        predictions: bool, optional (default=False)
            When set to True, the predictions of all the models models are returned as dataframe.

        sort_by: string, optional (default='RMSE')
            Sort models by a metric. Available options are 'RMSE', 'MAE', 'MPL', 'MPE', 'MAPE',
            'R-Squared', 'Adjusted R-Squared' or a custom metric identified by its name and
            provided by custom_metric.

        random_state: int, optional (default=42)
            Reproducibiility seed.

        estimators: list, optional (default='all')
            list of Estimators (regression algorithms) names or just 'all' (default='all')

        preprocess: bool, preprocessing is done when set to True

        n_layers: int, optional (default=1)
            Number of layers in the network. When set to 1, the model is equivalent to a MTS.

        h: int, optional (default=None)
            Number of steps ahead to predict (when used, must be > 0 and < X_test.shape[0]).

        All the other parameters are the same as MTS's.
    
    Attributes:

        models_: dict-object
            Returns a dictionary with each model pipeline as value
            with key as name of models.
        
        best_model_: object
            Returns the best model pipeline based on the sort_by metric.

    Examples:

        See https://thierrymoudiki.github.io/blog/2023/10/29/python/quasirandomizednn/MTS-LazyPredict

    """

    def __init__(
        self,
        verbose=0,
        ignore_warnings=True,
        custom_metric=None,
        predictions=False,
        sort_by=None, # leave it as is 
        random_state=42,
        estimators="all",
        preprocess=False,
        n_layers=1,
        h=None,
        # MTS attributes
        obj=None,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        lags=15,
        type_pi="scp2-kde",
        block_size=None,
        replications=None,
        kernel=None,
        agg="mean",
        seed=123,
        backend="cpu",
        show_progress=False,
    ):
        self.verbose = verbose
        self.ignore_warnings = ignore_warnings
        self.custom_metric = custom_metric
        self.predictions = predictions
        self.sort_by = sort_by
        self.models_ = {}
        self.best_model_ = None
        self.random_state = random_state
        self.estimators = estimators
        self.preprocess = preprocess
        self.n_layers = n_layers
        self.h = h
        super().__init__(
            obj=obj,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            seed=seed,
            backend=backend,
            lags=lags,
            type_pi=type_pi,
            block_size=block_size,
            replications=replications,
            kernel=kernel,
            agg=agg,
            verbose=verbose,
            show_progress=show_progress,
        )
        if self.replications is not None or self.type_pi == "gaussian":
            if self.sort_by is None:
                self.sort_by = "WINKLERSCORE"
        else:
            if self.sort_by is None:
                self.sort_by = "RMSE"

    def fit(self, X_train, X_test, xreg=None, per_series=False, **kwargs):
        """Fit Regression algorithms to X_train, predict and score on X_test.

        Parameters:

            X_train: array-like or data frame,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            X_test: array-like or data frame,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

            xreg: array-like, optional (default=None)
                Additional (external) regressors to be passed to self.obj
                xreg must be in 'increasing' order (most recent observations last)

            per_series: bool, optional (default=False)
                When set to True, the metrics are computed series by series.

            **kwargs: dict, optional (default=None)
                Additional parameters to be passed to `fit` method of `obj`.

        Returns:

            scores: Pandas DataFrame
                Returns metrics of all the models in a Pandas DataFrame.

            predictions: Pandas DataFrame
                Returns predictions of all the models in a Pandas DataFrame.

        """
        R2 = []
        ADJR2 = []
        ME = []
        MPL = []
        RMSE = []
        MAE = []
        MPE = []
        MAPE = []
        WINKLERSCORE = []
        COVERAGE = []

        # WIN = []
        names = []
        TIME = []
        predictions = {}

        if self.custom_metric is not None:
            CUSTOM_METRIC = []

        if self.h is None:
            assert X_test is not None, "If h is None, X_test must be provided."

        if isinstance(X_train, np.ndarray):
            X_train = pd.DataFrame(X_train)
            X_test = pd.DataFrame(X_test)

        self.series_names = X_train.columns.tolist()

        X_train = convert_df_to_numeric(X_train)
        X_test = convert_df_to_numeric(X_test)

        numeric_features = X_train.select_dtypes(include=[np.number]).columns
        categorical_features = X_train.select_dtypes(include=["object"]).columns

        categorical_low, categorical_high = get_card_split(
            X_train, categorical_features
        )

        if self.preprocess:
            preprocessor = ColumnTransformer(
                transformers=[
                    ("numeric", numeric_transformer, numeric_features),
                    (
                        "categorical_low",
                        categorical_transformer_low,
                        categorical_low,
                    ),
                    (
                        "categorical_high",
                        categorical_transformer_high,
                        categorical_high,
                    ),
                ]
            )

        # baselines (Classical MTS) ----
        for i, name in enumerate(["ARIMA", "ETS", "Theta", "VAR", "VECM"]):           
            try: 
                start = time.time()
                regr = ClassicalMTS(model=name)
                regr.fit(X_train, **kwargs)
                self.models_[name] = regr
                if self.h is None:
                    X_pred = regr.predict(h=X_test.shape[0], **kwargs)
                else:
                    assert self.h > 0, "h must be > 0"
                    X_pred = regr.predict(h=self.h, **kwargs)
                    try:
                        X_test = X_test[0: self.h, :]
                    except Exception as e:
                        X_test = X_test.iloc[0: self.h, :]

                if per_series == False:
                    rmse = np.sqrt(np.mean((X_test - X_pred.mean)**2))
                    mae = mean_absolute_error(X_test, X_pred.mean)
                    mpl = mean_pinball_loss(X_test, X_pred.mean)
                else:
                    rmse = mean_errors(
                        actual=X_test,
                        pred=X_pred,
                        scoring="root_mean_squared_error",
                        per_series=True,
                    )
                    mae = mean_errors(
                        actual=X_test,
                        pred=X_pred,
                        scoring="mean_absolute_error",
                        per_series=True,
                    )
                    mpl = mean_errors(
                        actual=X_test,
                        pred=X_pred,
                        scoring="mean_pinball_loss",
                        per_series=True,
                    )
            except Exception as exception:
                continue

            names.append(name)
            RMSE.append(rmse)
            MAE.append(mae)
            MPL.append(mpl)

            if self.custom_metric is not None:                        
                try: 
                    if self.h is None:
                        custom_metric = self.custom_metric(X_test, X_pred)
                    else:
                        custom_metric = self.custom_metric(X_test_h, X_pred)
                    CUSTOM_METRIC.append(custom_metric)
                except Exception as e:
                    custom_metric = np.iinfo(np.float32).max
                    CUSTOM_METRIC.append(np.iinfo(np.float32).max)

            if (self.replications is not None) or (self.type_pi == "gaussian"):
                if per_series == False:
                    winklerscore = winkler_score(
                        obj=X_pred, actual=X_test, level=95
                    )
                    coveragecalc = coverage(X_pred, X_test, level=95)
                else: 
                    winklerscore = winkler_score(
                        obj=X_pred, actual=X_test, level=95, per_series=True
                    )
                    coveragecalc = coverage(X_pred, X_test, level=95, per_series=True)
                WINKLERSCORE.append(winklerscore)
                COVERAGE.append(coveragecalc)
            TIME.append(time.time() - start)        
        
        if self.estimators == "all":
            if self.n_layers <= 1:
                self.regressors = REGRESSORSMTS
            else:
                self.regressors = DEEPREGRESSORSMTS
        else:
            if self.n_layers <= 1:
                self.regressors = [
                    ("MTS(" + est[0] + ")", est[1])
                    for est in all_estimators()
                    if (
                        issubclass(est[1], RegressorMixin)
                        and (est[0] in self.estimators)
                    )
                ]
            else:  # self.n_layers > 1
                self.regressors = [
                    ("DeepMTS(" + est[0] + ")", est[1])
                    for est in all_estimators()
                    if (
                        issubclass(est[1], RegressorMixin)
                        and (est[0] in self.estimators)
                    )
                ]

        if self.preprocess is True:
            for name, model in tqdm(self.regressors):  # do parallel exec
                start = time.time()
                try:
                    if "random_state" in model().get_params().keys():
                        pipe = Pipeline(
                            steps=[
                                ("preprocessor", preprocessor),
                                (
                                    "regressor",
                                    DeepMTS(
                                        obj=model(
                                            random_state=self.random_state,
                                            **kwargs
                                        ),
                                        n_layers=self.n_layers,
                                        n_hidden_features=self.n_hidden_features,
                                        activation_name=self.activation_name,
                                        a=self.a,
                                        nodes_sim=self.nodes_sim,
                                        bias=self.bias,
                                        dropout=self.dropout,
                                        direct_link=self.direct_link,
                                        n_clusters=self.n_clusters,
                                        cluster_encode=self.cluster_encode,
                                        type_clust=self.type_clust,
                                        type_scaling=self.type_scaling,
                                        lags=self.lags,
                                        type_pi=self.type_pi,
                                        block_size=self.block_size,
                                        replications=self.replications,
                                        kernel=self.kernel,
                                        agg=self.agg,
                                        seed=self.seed,
                                        backend=self.backend,
                                        show_progress=self.show_progress,
                                    ),
                                ),
                            ]
                        )
                    else:  # "random_state" in model().get_params().keys()
                        pipe = Pipeline(
                            steps=[
                                ("preprocessor", preprocessor),
                                (
                                    "regressor",
                                    DeepMTS(
                                        obj=model(**kwargs),
                                        n_layers=self.n_layers,
                                        n_hidden_features=self.n_hidden_features,
                                        activation_name=self.activation_name,
                                        a=self.a,
                                        nodes_sim=self.nodes_sim,
                                        bias=self.bias,
                                        dropout=self.dropout,
                                        direct_link=self.direct_link,
                                        n_clusters=self.n_clusters,
                                        cluster_encode=self.cluster_encode,
                                        type_clust=self.type_clust,
                                        type_scaling=self.type_scaling,
                                        lags=self.lags,
                                        type_pi=self.type_pi,
                                        block_size=self.block_size,
                                        replications=self.replications,
                                        kernel=self.kernel,
                                        agg=self.agg,
                                        seed=self.seed,
                                        backend=self.backend,
                                        show_progress=self.show_progress,
                                    ),
                                ),
                            ]
                        )

                    pipe.fit(X_train, **kwargs)
                    # pipe.fit(X_train, xreg=xreg)

                    self.models_[name] = pipe

                    if self.h is None:
                        X_pred = pipe["regressor"].predict(h=self.h, **kwargs)
                    else:
                        assert self.h > 0, "h must be > 0"
                        X_pred = pipe["regressor"].predict(h=self.h, **kwargs)

                    if (self.replications is not None) or (
                        self.type_pi == "gaussian"
                    ):
                        if per_series == False:
                            rmse = np.sqrt(np.mean((X_test - X_pred.mean)**2))
                            mae = mean_absolute_error(X_test, X_pred.mean)
                            mpl = mean_pinball_loss(X_test, X_pred.mean)
                            winklerscore = winkler_score(
                                obj=X_pred, actual=X_test, level=95
                            )
                            coveragecalc = coverage(X_pred, X_test, level=95)
                        else:
                            rmse = mean_errors(
                                actual=X_test,
                                pred=X_pred,
                                scoring="root_mean_squared_error",
                                per_series=True,
                            )
                            mae = mean_errors(
                                actual=X_test,
                                pred=X_pred,
                                scoring="mean_absolute_error",
                                per_series=True,
                            )
                            mpl = mean_errors(
                                actual=X_test,
                                pred=X_pred,
                                scoring="mean_pinball_loss",
                                per_series=True,
                            )
                            winklerscore = winkler_score(
                                obj=X_pred,
                                actual=X_test,
                                level=95,
                                per_series=True,
                            )
                            coveragecalc = coverage(
                                X_pred, X_test, level=95, per_series=True
                            )
                    else:
                        if per_series == False:
                            rmse = np.sqrt(np.mean((X_test - X_pred)**2))
                            mae = mean_absolute_error(X_test, X_pred)
                            mpl = mean_pinball_loss(X_test, X_pred)
                        else:
                            rmse = mean_errors(
                                actual=X_test,
                                pred=X_pred,
                                scoring="root_mean_squared_error",
                                per_series=True,
                            )
                            mae = mean_errors(
                                actual=X_test,
                                pred=X_pred,
                                scoring="mean_absolute_error",
                                per_series=True,
                            )
                            mpl = mean_errors(
                                actual=X_test,
                                pred=X_pred,
                                scoring="mean_pinball_loss",
                                per_series=True,
                            )

                    names.append(name)
                    RMSE.append(rmse)
                    MAE.append(mae)
                    MPL.append(mpl)

                    if (self.replications is not None) or (
                        self.type_pi == "gaussian"
                    ):
                        WINKLERSCORE.append(winklerscore)
                        COVERAGE.append(coveragecalc)
                    TIME.append(time.time() - start)

                    if self.custom_metric is not None:                        
                        try: 
                            custom_metric = self.custom_metric(X_test, X_pred)
                            CUSTOM_METRIC.append(custom_metric)
                        except Exception as e:
                            custom_metric = np.iinfo(np.float32).max
                            CUSTOM_METRIC.append(custom_metric)

                    if self.verbose > 0:
                        if (self.replications is not None) or (
                            self.type_pi == "gaussian"
                        ):
                            scores_verbose = {
                                "Model": name,
                                "RMSE": rmse,
                                "MAE": mae,
                                "MPL": mpl,
                                "WINKLERSCORE": winklerscore,
                                "COVERAGE": coveragecalc,
                                "Time taken": time.time() - start,
                            }
                        else:
                            scores_verbose = {
                                "Model": name,
                                "RMSE": rmse,
                                "MAE": mae,
                                "MPL": mpl,
                                "Time taken": time.time() - start,
                            }

                        if self.custom_metric is not None:
                            scores_verbose["Custom metric"] = custom_metric
                            
                    if self.predictions:
                        predictions[name] = X_pred
                except Exception as exception:
                    if self.ignore_warnings is False:
                        print(name + " model failed to execute")
                        print(exception)

        else:  # no preprocessing

            for name, model in tqdm(self.regressors):  # do parallel exec
                start = time.time()
                try:
                    if "random_state" in model().get_params().keys():
                        pipe = DeepMTS(
                            obj=model(random_state=self.random_state, **kwargs),
                            n_layers=self.n_layers,
                            n_hidden_features=self.n_hidden_features,
                            activation_name=self.activation_name,
                            a=self.a,
                            nodes_sim=self.nodes_sim,
                            bias=self.bias,
                            dropout=self.dropout,
                            direct_link=self.direct_link,
                            n_clusters=self.n_clusters,
                            cluster_encode=self.cluster_encode,
                            type_clust=self.type_clust,
                            type_scaling=self.type_scaling,
                            lags=self.lags,
                            type_pi=self.type_pi,
                            block_size=self.block_size,
                            replications=self.replications,
                            kernel=self.kernel,
                            agg=self.agg,
                            seed=self.seed,
                            backend=self.backend,
                            show_progress=self.show_progress,
                        )
                    else:
                        pipe = DeepMTS(
                            obj=model(**kwargs),
                            n_layers=self.n_layers,
                            n_hidden_features=self.n_hidden_features,
                            activation_name=self.activation_name,
                            a=self.a,
                            nodes_sim=self.nodes_sim,
                            bias=self.bias,
                            dropout=self.dropout,
                            direct_link=self.direct_link,
                            n_clusters=self.n_clusters,
                            cluster_encode=self.cluster_encode,
                            type_clust=self.type_clust,
                            type_scaling=self.type_scaling,
                            lags=self.lags,
                            type_pi=self.type_pi,
                            block_size=self.block_size,
                            replications=self.replications,
                            kernel=self.kernel,
                            agg=self.agg,
                            seed=self.seed,
                            backend=self.backend,
                            show_progress=self.show_progress,
                        )

                    pipe.fit(X_train, xreg, **kwargs)
                    # pipe.fit(X_train, xreg=xreg) # DO xreg like in `ahead`

                    self.models_[name] = pipe

                    if self.preprocess is True:
                        if self.h is None:
                            X_pred = pipe["regressor"].predict(
                                h=X_test.shape[0], **kwargs
                            )
                        else:
                            assert (
                                self.h > 0 and self.h <= X_test.shape[0]
                            ), "h must be > 0 and < X_test.shape[0]"
                            X_pred = pipe["regressor"].predict(
                                h=self.h, **kwargs
                            )

                    else:

                        if self.h is None:
                            X_pred = pipe.predict(
                                h=X_test.shape[0], **kwargs
                            )  # X_pred = pipe.predict(h=X_test.shape[0], new_xreg=new_xreg) ## DO xreg like in `ahead`
                        else:
                            assert (
                                self.h > 0 and self.h <= X_test.shape[0]
                            ), "h must be > 0 and < X_test.shape[0]"
                            X_pred = pipe.predict(h=self.h, **kwargs)

                    if self.h is None:
                        if (self.replications is not None) or (
                            self.type_pi == "gaussian"
                        ):

                            if per_series == True:
                                rmse = mean_errors(
                                    actual=X_test,
                                    pred=X_pred.mean,
                                    scoring="root_mean_squared_error",
                                    per_series=True,
                                )
                                mae = mean_errors(
                                    actual=X_test,
                                    pred=X_pred.mean,
                                    scoring="mean_absolute_error",
                                    per_series=True,
                                )
                                mpl = mean_errors(
                                    actual=X_test,
                                    pred=X_pred.mean,
                                    scoring="mean_pinball_loss",
                                    per_series=True,
                                )
                                winklerscore = winkler_score(
                                    obj=X_pred,
                                    actual=X_test,
                                    level=95,
                                    per_series=True,
                                )
                                coveragecalc = coverage(
                                    X_pred, X_test, level=95, per_series=True
                                )
                            else:
                                rmse = np.sqrt(np.mean((X_test - X_pred.mean)**2))
                                mae = mean_absolute_error(X_test, X_pred.mean)
                                mpl = mean_pinball_loss(X_test, X_pred.mean)
                                winklerscore = winkler_score(
                                    obj=X_pred, actual=X_test, level=95
                                )
                                coveragecalc = coverage(
                                    X_pred, X_test, level=95
                                )
                        else:  # no prediction interval
                            if per_series == True:
                                rmse = mean_errors(
                                    actual=X_test,
                                    pred=X_pred,
                                    scoring="root_mean_squared_error",
                                    per_series=True,
                                )
                                mae = mean_errors(
                                    actual=X_test,
                                    pred=X_pred,
                                    scoring="mean_absolute_error",
                                    per_series=True,
                                )
                                mpl = mean_errors(
                                    actual=X_test,
                                    pred=X_pred,
                                    scoring="mean_pinball_loss",
                                    per_series=True,
                                )
                            else:
                                rmse = np.sqrt(np.mean((X_test - X_pred)**2))
                                mae = mean_absolute_error(X_test, X_pred)
                                mpl = mean_pinball_loss(X_test, X_pred)
                    else:  # self.h is not None
                        if (self.replications is not None) or (
                            self.type_pi == "gaussian"
                        ):

                            if per_series == False:
                                if isinstance(X_test, pd.DataFrame) == False:
                                    X_test_h = X_test[0: self.h, :]
                                    rmse = np.sqrt(np.mean((X_test_h - X_pred.mean)**2))
                                    mae = mean_absolute_error(
                                        X_test_h, X_pred.mean
                                    )
                                    mpl = mean_pinball_loss(
                                        X_test_h, X_pred.mean
                                    )
                                    winklerscore = winkler_score(
                                        obj=X_pred, actual=X_test_h, level=95
                                    )
                                    coveragecalc = coverage(
                                        X_pred, X_test_h, level=95
                                    )
                                else:
                                    X_test_h = X_test.iloc[0: self.h, :]
                                    rmse = np.sqrt(np.mean((X_test_h - X_pred.mean)**2))
                                    mae = mean_absolute_error(
                                        X_test_h, X_pred.mean
                                    )
                                    mpl = mean_pinball_loss(
                                        X_test_h, X_pred.mean
                                    )
                                    winklerscore = winkler_score(
                                        obj=X_pred, actual=X_test_h, level=95
                                    )
                                    coveragecalc = coverage(
                                        X_pred, X_test_h, level=95
                                    )
                            else:
                                if isinstance(X_test, pd.DataFrame):
                                    X_test_h = X_test.iloc[0: self.h, :]
                                    rmse = mean_errors(
                                        actual=X_test_h,
                                        pred=X_pred,
                                        scoring="root_mean_squared_error",
                                        per_series=True,
                                    )
                                    mae = mean_errors(
                                        actual=X_test_h,
                                        pred=X_pred,
                                        scoring="mean_absolute_error",
                                        per_series=True,
                                    )
                                    mpl = mean_errors(
                                        actual=X_test_h,
                                        pred=X_pred,
                                        scoring="mean_pinball_loss",
                                        per_series=True,
                                    )
                                    winklerscore = winkler_score(
                                        obj=X_pred,
                                        actual=X_test_h,
                                        level=95,
                                        per_series=True,
                                    )
                                    coveragecalc = coverage(
                                        X_pred,
                                        X_test_h,
                                        level=95,
                                        per_series=True,
                                    )
                                else:
                                    X_test_h = X_test[0: self.h, :]
                                    rmse = mean_errors(
                                        actual=X_test_h,
                                        pred=X_pred,
                                        scoring="root_mean_squared_error",
                                        per_series=True,
                                    )
                                    mae = mean_errors(
                                        actual=X_test_h,
                                        pred=X_pred,
                                        scoring="mean_absolute_error",
                                        per_series=True,
                                    )
                                    mpl = mean_errors(
                                        actual=X_test_h,
                                        pred=X_pred,
                                        scoring="mean_pinball_loss",
                                        per_series=True,
                                    )
                                    winklerscore = winkler_score(
                                        obj=X_pred,
                                        actual=X_test_h,
                                        level=95,
                                        per_series=True,
                                    )
                                    coveragecalc = coverage(
                                        X_pred,
                                        X_test_h,
                                        level=95,
                                        per_series=True,
                                    )
                        else:  # no prediction interval

                            if per_series == False:
                                if isinstance(X_test, pd.DataFrame):
                                    X_test_h = X_test.iloc[0: self.h, :]
                                    rmse = np.sqrt(np.mean((X_test_h - X_pred)**2))
                                    mae = mean_absolute_error(X_test_h, X_pred)
                                    mpl = mean_pinball_loss(X_test_h, X_pred)
                                else:
                                    X_test_h = X_test[0: self.h, :]
                                    rmse = np.sqrt(np.mean((X_test_h - X_pred)**2))
                                    mae = mean_absolute_error(X_test_h, X_pred)
                                    mpl = mean_pinball_loss(X_test_h, X_pred)
                            else:
                                if isinstance(X_test, pd.DataFrame):
                                    X_test_h = X_test.iloc[0: self.h, :]
                                    rmse = mean_errors(
                                        actual=X_test_h,
                                        pred=X_pred,
                                        scoring="root_mean_squared_error",
                                        per_series=True,
                                    )
                                    mae = mean_errors(
                                        actual=X_test_h,
                                        pred=X_pred,
                                        scoring="mean_absolute_error",
                                        per_series=True,
                                    )
                                    mpl = mean_errors(
                                        actual=X_test_h,
                                        pred=X_pred,
                                        scoring="mean_pinball_loss",
                                        per_series=True,
                                    )
                                else:
                                    X_test_h = X_test[0: self.h, :]
                                    rmse = mean_errors(
                                        actual=X_test_h,
                                        pred=X_pred,
                                        scoring="root_mean_squared_error",
                                        per_series=True,
                                    )
                                    mae = mean_errors(
                                        actual=X_test_h,
                                        pred=X_pred,
                                        scoring="mean_absolute_error",
                                        per_series=True,
                                    )

                    names.append(name)
                    RMSE.append(rmse)
                    MAE.append(mae)
                    MPL.append(mpl)
                    if (self.replications is not None) or (
                        self.type_pi == "gaussian"
                    ):
                        WINKLERSCORE.append(winklerscore)
                        COVERAGE.append(coveragecalc)
                    TIME.append(time.time() - start)

                    if self.custom_metric is not None:                        
                        try: 
                            if self.h is None:
                                custom_metric = self.custom_metric(X_test, X_pred)
                            else:
                                custom_metric = self.custom_metric(X_test_h, X_pred)
                            CUSTOM_METRIC.append(custom_metric)
                        except Exception as e:
                            custom_metric = np.iinfo(np.float32).max
                            CUSTOM_METRIC.append(np.iinfo(np.float32).max)

                    if self.verbose > 0:
                        if (self.replications is not None) or (
                            self.type_pi == "gaussian"
                        ):
                            scores_verbose = {
                                "Model": name,
                                "RMSE": rmse,
                                "MAE": mae,
                                "MPL": mpl,
                                "WINKLERSCORE": winklerscore,
                                "COVERAGE": coveragecalc,
                                "Time taken": time.time() - start,
                            }
                        else:
                            scores_verbose = {
                                "Model": name,
                                "RMSE": rmse,
                                "MAE": mae,
                                "MPL": mpl,
                                "Time taken": time.time() - start,
                            }

                        if self.custom_metric is not None:
                            scores_verbose["Custom metric"] = custom_metric                            
                       
                    if self.predictions:
                        predictions[name] = X_pred

                except Exception as exception:
                    if self.ignore_warnings is False:
                        print(name + " model failed to execute")
                        print(exception)

        if (self.replications is not None) or (self.type_pi == "gaussian"):
            scores = {
                "Model": names,
                "RMSE": RMSE,
                "MAE": MAE,
                "MPL": MPL,
                "WINKLERSCORE": WINKLERSCORE,
                "COVERAGE": COVERAGE,
                "Time Taken": TIME,
            }
        else:
            scores = {
                "Model": names,
                "RMSE": RMSE,
                "MAE": MAE,
                "MPL": MPL,
                "Time Taken": TIME,
            }

        if self.custom_metric is not None:
            scores["Custom metric"] = CUSTOM_METRIC
        
        if per_series:
            scores = dict_to_dataframe_series(scores, self.series_names)
        else:
            scores = pd.DataFrame(scores)
       
        try: # case per_series, can't be sorted
            scores = scores.sort_values(by=self.sort_by, ascending=True).set_index("Model")
            
            self.best_model_ = self.models_[scores.index[0]]
        except Exception as e:
            pass 

        if self.predictions is True:

            return scores, predictions
        
        return scores

    def get_best_model(self):
        """
        This function returns the best model pipeline based on the sort_by metric.

        Returns:

            best_model: object,
                Returns the best model pipeline based on the sort_by metric.

        """
        return self.best_model_

    def provide_models(self, X_train, X_test):
        """
        This function returns all the model objects trained in fit function.
        If fit is not called already, then we call fit and then return the models.

        Parameters:

            X_train : array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            X_test : array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

        Returns:

            models: dict-object,
                Returns a dictionary with each model pipeline as value
                with key as name of models.

        """
        if self.h is None:
            if len(self.models_.keys()) == 0:
                self.fit(X_train, X_test)
        else:
            if len(self.models_.keys()) == 0:
                if isinstance(X_test, pd.DataFrame):
                    self.fit(X_train, X_test.iloc[0: self.h, :])
                else:
                    self.fit(X_train, X_test[0: self.h, :])

        return self.models_

class LazyMTS(LazyDeepMTS):
    """
    Fitting -- almost -- all the regression algorithms to multivariate time series
    and returning their scores (no layers).

    Parameters:

        verbose: int, optional (default=0)
            Any positive number for verbosity.

        ignore_warnings: bool, optional (default=True)
            When set to True, the warning related to algorigms that are not
            able to run are ignored.

        custom_metric: function, optional (default=None)
            When function is provided, models are evaluated based on the custom
              evaluation metric provided.

        predictions: bool, optional (default=False)
            When set to True, the predictions of all the models models are returned as dataframe.

        sort_by: string, optional (default='RMSE')
            Sort models by a metric. Available options are 'RMSE', 'MAE', 'MPL', 'MPE', 'MAPE',
            'R-Squared', 'Adjusted R-Squared' or a custom metric identified by its name and
            provided by custom_metric.

        random_state: int, optional (default=42)
            Reproducibiility seed.

        estimators: list, optional (default='all')
            list of Estimators (regression algorithms) names or just 'all' (default='all')

        preprocess: bool, preprocessing is done when set to True

        h: int, optional (default=None)
            Number of steps ahead to predict (when used, must be > 0 and < X_test.shape[0]).

        All the other parameters are the same as MTS's.
    
    Attributes:

        models_: dict-object
            Returns a dictionary with each model pipeline as value
            with key as name of models.
        
        best_model_: object
            Returns the best model pipeline based on the sort_by metric.

    Examples:

        See https://thierrymoudiki.github.io/blog/2023/10/29/python/quasirandomizednn/MTS-LazyPredict

    """

    def __init__(
        self,
        verbose=0,
        ignore_warnings=True,
        custom_metric=None,
        predictions=False,
        sort_by=None, # leave it as is 
        random_state=42,
        estimators="all",
        preprocess=False,
        h=None,
        # MTS attributes
        obj=None,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        lags=15,
        type_pi="scp2-kde",
        block_size=None,
        replications=None,
        kernel=None,
        agg="mean",
        seed=123,
        backend="cpu",
        show_progress=False,
    ):
        super().__init__(
            verbose=verbose,
            ignore_warnings=ignore_warnings,
            custom_metric=custom_metric,
            predictions=predictions,
            sort_by=sort_by,
            random_state=random_state,
            estimators=estimators,
            preprocess=preprocess,
            n_layers=1,
            h=h,
            obj=obj,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            lags=lags,
            type_pi=type_pi,
            block_size=block_size,
            replications=replications,
            kernel=kernel,
            agg=agg,
            seed=seed,
            backend=backend,
            show_progress=show_progress,
        )



# Authors: Thierry Moudiki
#
# License: BSD 3 Clear Clause

import copy
import numpy as np
import pandas as pd
import sklearn.metrics as skm2
import matplotlib.pyplot as plt
from collections import namedtuple
from copy import deepcopy
from functools import partial
from scipy.stats import describe, norm
from sklearn.metrics import (
    mean_absolute_error,
    mean_absolute_percentage_error,
)
from sklearn.neighbors import KernelDensity
from sklearn.model_selection import GridSearchCV
try: 
    from statsmodels.tsa.api import VAR
    from statsmodels.tsa.vector_ar.vecm import VECM
    from statsmodels.tsa.arima.model import ARIMA
    from statsmodels.tsa.holtwinters import ExponentialSmoothing
    from statsmodels.tsa.forecasting.theta import ThetaModel
except ImportError:
    pass

from tqdm import tqdm
from ..base import Base
from ..sampling import vinecopula_sample
from ..simulation import getsims
from ..utils import matrixops as mo
from ..utils import misc as mx
from ..utils import timeseries as ts
from ..utils import convert_df_to_numeric, coverage, winkler_score, mean_errors
from ..utils import TimeSeriesSplit


class ClassicalMTS(Base):
    """Multivariate time series (FactorMTS) forecasting with Factor models

    Parameters:

        model: type of model: str.
            currently, 'VAR', 'VECM', 'ARIMA', 'ETS', 'Theta'

    Attributes:

        df_: data frame
            the input data frame, in case a data.frame is provided to `fit`

        level_: int
            level of confidence for prediction intervals (default is 95)

    Examples:
    See examples/classical_mts_timeseries.py
    """

    # construct the object -----

    def __init__(self, model="VAR"):

        self.model = model
        if self.model == "VAR":
            self.obj = VAR
        elif self.model == "VECM":
            self.obj = VECM
        elif self.model == "ARIMA":
            self.obj = ARIMA
        elif self.model == "ETS":
            self.obj = ExponentialSmoothing
        elif self.model == "Theta":
            self.obj = ThetaModel
        else:
            raise ValueError("model not recognized")
        self.n_series = None
        self.replications = None
        self.mean_ = None
        self.upper_ = None
        self.lower_ = None
        self.output_dates_ = None
        self.alpha_ = None
        self.df_ = None
        self.residuals_ = []
        self.sims_ = None
        self.level_ = None

    def fit(self, X, **kwargs):
        """Fit FactorMTS model to training data X, with optional regressors xreg

        Parameters:

        X: {array-like}, shape = [n_samples, n_features]
            Training time series, where n_samples is the number
            of samples and n_features is the number of features;
            X must be in increasing order (most recent observations last)

        **kwargs: for now, additional parameters to be passed to for kernel density estimation, when needed (see sklearn.neighbors.KernelDensity)

        Returns:

        self: object
        """

        try:
            self.n_series = X.shape[1]        
        except Exception:
            self.n_series = 1

        if (
            isinstance(X, pd.DataFrame) is False
        ) and isinstance(X, pd.Series) is False :  # input data set is a numpy array
            
            X = pd.DataFrame(X)
            if self.n_series > 1: 
                self.series_names = ["series" + str(i) for i in range(X.shape[1])]
            else: 
                self.series_names = "series0"

        else:  # input data set is a DataFrame or Series with column names
            
            X_index = None
            if X.index is not None and len(X.shape) > 1:
                X_index = X.index
                X = copy.deepcopy(mo.convert_df_to_numeric(X))
            if X_index is not None:
                try:
                    X.index = X_index
                except Exception:
                    pass
            if isinstance(X, pd.DataFrame): 
                self.series_names = X.columns.tolist()
            else:
                self.series_names = X.name

        if isinstance(X, pd.DataFrame) or isinstance(X, pd.Series):
            self.df_ = X
            X = X.values
            self.df_.columns = self.series_names
            self.input_dates = ts.compute_input_dates(self.df_)
        else:
            self.df_ = pd.DataFrame(X, columns=self.series_names)            

        if self.model == "Theta":
            self.obj = self.obj(self.df_, 
                                **kwargs).fit()
        else:        
            self.obj = self.obj(X, **kwargs).fit(**kwargs)        

        return self

    def predict(self, h=5, level=95, **kwargs):
        """Forecast all the time series, h steps ahead

        Parameters:

        h: {integer}
            Forecasting horizon

        **kwargs: additional parameters to be passed to
                self.cook_test_set

        Returns:

        model predictions for horizon = h: {array-like}

        """

        self.output_dates_, frequency = ts.compute_output_dates(self.df_, h)

        self.level_ = level

        self.lower_ = None  # do not remove (/!\)

        self.upper_ = None  # do not remove (/!\)

        self.sims_ = None  # do not remove (/!\)

        self.level_ = level

        self.alpha_ = 100 - level
        
        pi_multiplier = norm.ppf(1 - self.alpha_ / 200)

        # Named tuple for forecast results
        DescribeResult = namedtuple(
            "DescribeResult", ("mean", "lower", "upper")
        )

        if self.model == "VAR":
            mean_forecast, lower_bound, upper_bound = (
                self.obj.forecast_interval(
                    self.obj.endog, steps=h, alpha=self.alpha_ / 100, **kwargs
                )
            )

        elif self.model == "VECM":
            forecast_result = self.obj.predict(steps=h)
            mean_forecast = forecast_result
            lower_bound, upper_bound = self._compute_confidence_intervals(
                forecast_result, alpha=self.alpha_ / 100, **kwargs
            )
        
        elif self.model == "ARIMA":
            forecast_result = self.obj.get_forecast(steps=h)
            mean_forecast = forecast_result.predicted_mean
            lower_bound = forecast_result.conf_int()[:, 0]
            upper_bound = forecast_result.conf_int()[:, 1]    
        
        elif self.model == "ETS":
            forecast_result = self.obj.forecast(steps=h)
            residuals = self.obj.resid
            std_errors = np.std(residuals)
            mean_forecast = forecast_result
            lower_bound = forecast_result - pi_multiplier * std_errors
            upper_bound = forecast_result + pi_multiplier * std_errors
        
        elif self.model == "Theta":
            try:
                mean_forecast = self.obj.forecast(steps=h).values 
                forecast_result = self.obj.prediction_intervals(steps=h, alpha=self.alpha_/100, 
                                                            **kwargs)            
                lower_bound = forecast_result["lower"].values 
                upper_bound = forecast_result["upper"].values 
            except Exception: 
                mean_forecast = self.obj.forecast(steps=h)
                forecast_result = self.obj.prediction_intervals(steps=h, alpha=self.alpha_/100, 
                                                            **kwargs)            
                lower_bound = forecast_result["lower"]
                upper_bound = forecast_result["upper"]
            
        else:

            raise ValueError("model not recognized")                    

        try: 
            self.mean_ = pd.DataFrame(mean_forecast, 
                                        columns=self.series_names, 
                                        index=self.output_dates_)        
            self.lower_ = pd.DataFrame(lower_bound, 
                                    columns=self.series_names, 
                                        index=self.output_dates_)
            self.upper_ = pd.DataFrame(upper_bound, 
                                    columns=self.series_names, 
                                        index=self.output_dates_)
        except Exception:
            self.mean_ = pd.Series(mean_forecast, 
                                    name=self.series_names, 
                                    index=self.output_dates_)        
            self.lower_ = pd.Series(lower_bound, 
                                    name=self.series_names, 
                                    index=self.output_dates_)
            self.upper_ = pd.Series(upper_bound, 
                                    name=self.series_names, 
                                    index=self.output_dates_)
        

        return DescribeResult(
            mean=self.mean_, lower=self.lower_, upper=self.upper_
        )

    def _compute_confidence_intervals(self, forecast_result, alpha):
        """
        Compute confidence intervals for VECM forecasts.
        Uses the covariance of residuals to approximate the confidence intervals.
        """
        residuals = self.obj.resid
        cov_matrix = np.cov(residuals.T)  # Covariance matrix of residuals
        std_errors = np.sqrt(np.diag(cov_matrix))  # Standard errors

        z_value = norm.ppf(1 - alpha / 2)  # Z-score for the given alpha level
        lower_bound = forecast_result - z_value * std_errors
        upper_bound = forecast_result + z_value * std_errors

        return lower_bound, upper_bound

    def score(self, X, training_index, testing_index, scoring=None, **kwargs):
        """Train on training_index, score on testing_index."""

        assert (
            bool(set(training_index).intersection(set(testing_index))) == False
        ), "Non-overlapping 'training_index' and 'testing_index' required"

        # Dimensions
        try:
            # multivariate time series
            n, p = X.shape
        except:
            # univariate time series
            n = X.shape[0]
            p = 1

        # Training and testing sets
        if p > 1:
            X_train = X[training_index, :]
            X_test = X[testing_index, :]
        else:
            X_train = X[training_index]
            X_test = X[testing_index]

        # Horizon
        h = len(testing_index)
        assert (
            len(training_index) + h
        ) <= n, "Please check lengths of training and testing windows"

        # Fit and predict
        self.fit(X_train, **kwargs)
        preds = self.predict(h=h, **kwargs)

        if scoring is None:
            scoring = "neg_root_mean_squared_error"

        # check inputs
        assert scoring in (
            "explained_variance",
            "neg_mean_absolute_error",
            "neg_mean_squared_error",
            "neg_root_mean_squared_error",
            "neg_mean_squared_log_error",
            "neg_median_absolute_error",
            "r2",
        ), "'scoring' should be in ('explained_variance', 'neg_mean_absolute_error', \
                               'neg_mean_squared_error', 'neg_root_mean_squared_error', 'neg_mean_squared_log_error', \
                               'neg_median_absolute_error', 'r2')"

        scoring_options = {
            "explained_variance": skm2.explained_variance_score,
            "neg_mean_absolute_error": skm2.mean_absolute_error,
            "neg_mean_squared_error": lambda x, y: np.mean((x - y)**2),
            "neg_root_mean_squared_error": lambda x, y: np.sqrt(np.mean((x - y)**2)),
            "neg_mean_squared_log_error": skm2.mean_squared_log_error,
            "neg_median_absolute_error": skm2.median_absolute_error,
            "r2": skm2.r2_score,
        }

        # if p > 1:
        #     return tuple(
        #         [
        #             scoring_options[scoring](
        #                 X_test[:, i], preds[:, i]#, **kwargs
        #             )
        #             for i in range(p)
        #         ]
        #     )
        # else:
        return scoring_options[scoring](X_test, preds)

    def plot(self, series=None, type_axis="dates", type_plot="pi"):
        """Plot time series forecast

        Parameters:

        series: {integer} or {string}
            series index or name

        """

        assert all(
            [
                self.mean_ is not None,
                self.lower_ is not None,
                self.upper_ is not None,
                self.output_dates_ is not None,
            ]
        ), "model forecasting must be obtained first (with predict)"

        if series is None:
            assert (
                self.n_series == 1
            ), "please specify series index or name (n_series > 1)"
            series = 0

        if isinstance(series, str):
            assert (
                series in self.series_names
            ), f"series {series} doesn't exist in the input dataset"
            series_idx = self.df_.columns.get_loc(series)
        else:
            assert isinstance(series, int) and (
                0 <= series < self.n_series
            ), f"check series index (< {self.n_series})"
            series_idx = series

        if isinstance(self.df_, pd.DataFrame): 
            y_all = list(self.df_.iloc[:, series_idx]) + list(
                self.mean_.iloc[:, series_idx]
            )
            y_test = list(self.mean_.iloc[:, series_idx])
        else:
            y_all = list(self.df_.values) + list(
                self.mean_.values
            )
            y_test = list(self.mean_.values)
        n_points_all = len(y_all)
        n_points_train = self.df_.shape[0]

        if type_axis == "numeric":
            x_all = [i for i in range(n_points_all)]
            x_test = [i for i in range(n_points_train, n_points_all)]

        if type_axis == "dates":  # use dates
            x_all = np.concatenate(
                (self.input_dates.values, self.output_dates_.values), axis=None
            )
            x_test = self.output_dates_.values

        if type_plot == "pi":
            fig, ax = plt.subplots()
            ax.plot(x_all, y_all, "-")
            ax.plot(x_test, y_test, "-", color="orange")
            try: 
                ax.fill_between(
                    x_test,
                    self.lower_.iloc[:, series_idx],
                    self.upper_.iloc[:, series_idx],
                    alpha=0.2,
                    color="orange",
                )
            except Exception:
                ax.fill_between(
                    x_test,
                    self.lower_.values,
                    self.upper_.values,
                    alpha=0.2,
                    color="orange",
                )
            if self.replications is None:
                if self.n_series > 1:
                    plt.title(
                        f"prediction intervals for {series}",
                        loc="left",
                        fontsize=12,
                        fontweight=0,
                        color="black",
                    )
                else:
                    plt.title(
                        f"prediction intervals for input time series",
                        loc="left",
                        fontsize=12,
                        fontweight=0,
                        color="black",
                    )
                plt.show()
            else:  # self.replications is not None
                if self.n_series > 1:
                    plt.title(
                        f"prediction intervals for {self.replications} simulations of {series}",
                        loc="left",
                        fontsize=12,
                        fontweight=0,
                        color="black",
                    )
                else:
                    plt.title(
                        f"prediction intervals for {self.replications} simulations of input time series",
                        loc="left",
                        fontsize=12,
                        fontweight=0,
                        color="black",
                    )
                plt.show()

        if type_plot == "spaghetti":
            palette = plt.get_cmap("Set1")
            sims_ix = getsims(self.sims_, series_idx)
            plt.plot(x_all, y_all, "-")
            for col_ix in range(
                sims_ix.shape[1]
            ):  # avoid this when there are thousands of simulations
                plt.plot(
                    x_test,
                    sims_ix[:, col_ix],
                    "-",
                    color=palette(col_ix),
                    linewidth=1,
                    alpha=0.9,
                )
            plt.plot(x_all, y_all, "-", color="black")
            plt.plot(x_test, y_test, "-", color="blue")
            # Add titles
            if self.n_series > 1:
                plt.title(
                    f"{self.replications} simulations of {series}",
                    loc="left",
                    fontsize=12,
                    fontweight=0,
                    color="black",
                )
            else:
                plt.title(
                    f"{self.replications} simulations of input time series",
                    loc="left",
                    fontsize=12,
                    fontweight=0,
                    color="black",
                )
            plt.xlabel("Time")
            plt.ylabel("Values")
            # Show the graph
            plt.show()

    def cross_val_score(
        self,
        X,
        scoring="root_mean_squared_error",
        n_jobs=None,
        verbose=0,
        xreg=None,
        initial_window=5,
        horizon=3,
        fixed_window=False,
        show_progress=True,
        level=95,
        **kwargs,
    ):
        """Evaluate a score by time series cross-validation.

        Parameters:

            X: {array-like, sparse matrix} of shape (n_samples, n_features)
                The data to fit.

            scoring: str or a function
                A str in ('root_mean_squared_error', 'mean_squared_error', 'mean_error',
                'mean_absolute_error', 'mean_error', 'mean_percentage_error',
                'mean_absolute_percentage_error',  'winkler_score', 'coverage')
                Or a function defined as 'coverage' and 'winkler_score' in `utils.timeseries`

            n_jobs: int, default=None
                Number of jobs to run in parallel.

            verbose: int, default=0
                The verbosity level.

            xreg: array-like, optional (default=None)
                Additional (external) regressors to be passed to `fit`
                xreg must be in 'increasing' order (most recent observations last)

            initial_window: int
                initial number of consecutive values in each training set sample

            horizon: int
                number of consecutive values in test set sample

            fixed_window: boolean
                if False, all training samples start at index 0, and the training
                window's size is increasing.
                if True, the training window's size is fixed, and the window is
                rolling forward

            show_progress: boolean
                if True, a progress bar is printed

            **kwargs: dict
                additional parameters to be passed to `fit` and `predict`

        Returns:

            A tuple: descriptive statistics or errors and raw errors

        """
        tscv = TimeSeriesSplit()

        tscv_obj = tscv.split(
            X,
            initial_window=initial_window,
            horizon=horizon,
            fixed_window=fixed_window,
        )

        if isinstance(scoring, str):

            assert scoring in (
                "root_mean_squared_error",
                "mean_squared_error",
                "mean_error",
                "mean_absolute_error",
                "mean_percentage_error",
                "mean_absolute_percentage_error",
                "winkler_score",
                "coverage",
            ), "must have scoring in ('root_mean_squared_error', 'mean_squared_error', 'mean_error', 'mean_absolute_error', 'mean_error', 'mean_percentage_error', 'mean_absolute_percentage_error',  'winkler_score', 'coverage')"

            def err_func(X_test, X_pred, scoring):
                if (self.replications is not None) or (
                    self.type_pi == "gaussian"
                ):  # probabilistic
                    if scoring == "winkler_score":
                        return winkler_score(X_pred, X_test, level=level)
                    elif scoring == "coverage":
                        return coverage(X_pred, X_test, level=level)
                    else:
                        return mean_errors(
                            pred=X_pred.mean, actual=X_test, scoring=scoring
                        )
                else:  # not probabilistic
                    return mean_errors(
                        pred=X_pred, actual=X_test, scoring=scoring
                    )

        else:  # isinstance(scoring, str) = False

            err_func = scoring

        errors = []

        train_indices = []

        test_indices = []

        for train_index, test_index in tscv_obj:
            train_indices.append(train_index)
            test_indices.append(test_index)

        if show_progress is True:
            iterator = tqdm(
                zip(train_indices, test_indices), total=len(train_indices)
            )
        else:
            iterator = zip(train_indices, test_indices)

        for train_index, test_index in iterator:

            if verbose == 1:
                print(f"TRAIN: {train_index}")
                print(f"TEST: {test_index}")

            if isinstance(X, pd.DataFrame):
                self.fit(X.iloc[train_index, :], xreg=xreg, **kwargs)
                X_test = X.iloc[test_index, :]
            else:
                self.fit(X[train_index, :], xreg=xreg, **kwargs)
                X_test = X[test_index, :]
            X_pred = self.predict(h=int(len(test_index)), level=level, **kwargs)

            errors.append(err_func(X_test, X_pred, scoring))

        res = np.asarray(errors)

        return res, describe(res)



# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
import pandas as pd
import sklearn.metrics as skm2
from .custom import Custom
from ..predictionset import PredictionSet
from ..utils import matrixops as mo
from sklearn.base import ClassifierMixin


class CustomClassifier(Custom, ClassifierMixin):
    """Custom Classification model

    Attributes:

        obj: object
            any object containing a method fit (obj.fit()) and a method predict
            (obj.predict())

        n_hidden_features: int
            number of nodes in the hidden layer

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton',
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or not
            (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out
            of the training

        direct_link: boolean
            indicates if the original predictors are included (True) in model''s
            fitting or not (False)

        n_clusters: int
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0:
                no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian
            Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant).
            Currently available: standardization ('std') or MinMax scaling ('minmax')

        col_sample: float
            percentage of covariates randomly chosen for training

        row_sample: float
            percentage of rows chosen for training, by stratified bootstrapping

        level: float
            confidence level for prediction sets. Default is None.

        pi_method: str
            method for constructing the prediction sets: 'icp', 'tcp' if level is not None. Default is 'icp'.

        seed: int
            reproducibility seed for nodes_sim=='uniform'

        backend: str
            "cpu" or "gpu" or "tpu"

    Examples:

    Note: it's better to use the `DeepClassifier` or `LazyDeepClassifier` classes directly

    ```python
    import nnetsauce as ns
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.model_selection import train_test_split
    from sklearn.datasets import load_digits
    from time import time

    digits = load_digits()
    X = digits.data
    y = digits.target
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2,
                                                        random_state=123)

    # layer 1 (base layer) ----
    layer1_regr = RandomForestClassifier(n_estimators=10, random_state=123)

    start = time()

    layer1_regr.fit(X_train, y_train)

    # Accuracy in layer 1
    print(layer1_regr.score(X_test, y_test))

    # layer 2 using layer 1 ----
    layer2_regr = ns.CustomClassifier(obj = layer1_regr, n_hidden_features=5,
                            direct_link=True, bias=True,
                            nodes_sim='uniform', activation_name='relu',
                            n_clusters=2, seed=123)
    layer2_regr.fit(X_train, y_train)

    # Accuracy in layer 2
    print(layer2_regr.score(X_test, y_test))

    # layer 3 using layer 2 ----
    layer3_regr = ns.CustomClassifier(obj = layer2_regr, n_hidden_features=10,
                            direct_link=True, bias=True, dropout=0.7,
                            nodes_sim='uniform', activation_name='relu',
                            n_clusters=2, seed=123)
    layer3_regr.fit(X_train, y_train)

    # Accuracy in layer 3
    print(layer3_regr.score(X_test, y_test))

    print(f"Elapsed {time() - start}")
    ```

    """

    # construct the object -----

    def __init__(
        self,
        obj,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        level=None,
        pi_method="icp",
        seed=123,
        backend="cpu",
    ):
        super().__init__(
            obj=obj,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
            backend=backend,
        )
        self.level = level
        self.pi_method = pi_method
        self.type_fit = "classification"
        if self.level is not None:
            self.obj = PredictionSet(
                self.obj, level=self.level, method=self.pi_method
            )

    def fit(self, X, y, sample_weight=None, **kwargs):
        """Fit custom model to training data (X, y).

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Target values.

            sample_weight: array-like, shape = [n_samples]
                Sample weights.

            **kwargs: additional parameters to be passed to
                        self.cook_training_set or self.obj.fit

        Returns:

            self: object
        """

        if len(X.shape) == 1:
            if isinstance(X, pd.DataFrame):
                X = pd.DataFrame(X.values.reshape(1, -1), columns=X.columns)    
            else:
                X = X.reshape(1, -1)

        output_y, scaled_Z = self.cook_training_set(y=y, X=X, **kwargs)
        self.classes_ =  np.unique(y)
        self.n_classes_ = len(self.classes_)  # for compatibility with         

        if self.level is not None:
            self.obj = PredictionSet(
                obj=self.obj, method=self.pi_method, level=self.level
            )

        # if sample_weights, else: (must use self.row_index)
        if sample_weight is not None:
            self.obj.fit(
                scaled_Z,
                output_y,
                sample_weight=sample_weight[self.index_row_].ravel(),
                # **kwargs
            )

            return self

        # if sample_weight is None:
        self.obj.fit(scaled_Z, output_y)
        self.classes_ = np.unique(y)  # for compatibility with sklearn
        self.n_classes_ = len(self.classes_)  # for compatibility with sklearn

        return self

    def partial_fit(self, X, y, sample_weight=None, **kwargs):
        """Partial fit custom model to training data (X, y).

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Subset of training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Subset of target values.

            sample_weight: array-like, shape = [n_samples]
                Sample weights.

            **kwargs: additional parameters to be passed to
                        self.cook_training_set or self.obj.fit

        Returns:

            self: object
        """

        if len(X.shape) == 1:
            if isinstance(X, pd.DataFrame):
                X = pd.DataFrame(X.values.reshape(1, -1), columns=X.columns)             
            else:
                X = X.reshape(1, -1)
            y = np.array([y], dtype=np.integer)

        output_y, scaled_Z = self.cook_training_set(y=y, X=X, **kwargs)
        self.n_classes_ = len(np.unique(y))  # for compatibility with sklearn

        # if sample_weights, else: (must use self.row_index)
        if sample_weight is not None:
            try:
                self.obj.partial_fit(
                    scaled_Z,
                    output_y,
                    sample_weight=sample_weight[self.index_row_].ravel(),
                    # **kwargs
                )
            except:
                NotImplementedError

            return self

        # if sample_weight is None:
        try:
            self.obj.partial_fit(scaled_Z, output_y)
        except:
            raise NotImplementedError

        self.classes_ = np.unique(y)  # for compatibility with sklearn
        self.n_classes_ = len(self.classes_)  # for compatibility with sklearn

        return self

    def predict(self, X, **kwargs):
        """Predict test data X.

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to
                    self.cook_test_set

        Returns:

            model predictions: {array-like}
        """

        if len(X.shape) == 1:
            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            return (
                self.obj.predict(self.cook_test_set(new_X, **kwargs), **kwargs)
            )[0]

        return self.obj.predict(self.cook_test_set(X, **kwargs), **kwargs)

    def predict_proba(self, X, **kwargs):
        """Predict probabilities for test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to
                    self.cook_test_set

        Returns:

            probability estimates for test data: {array-like}
        """

        if len(X.shape) == 1:
            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            return (
                self.obj.predict_proba(
                    self.cook_test_set(new_X, **kwargs), **kwargs
                )
            )[0]

        return self.obj.predict_proba(self.cook_test_set(X, **kwargs), **kwargs)



"""Tools for MTS."""

# Authors: Thierry Moudiki <thierry.moudiki@gmail.com>
#
# License: BSD 3 Clear

import numpy as np
import pandas as pd
try: 
    from sklearn.metrics import mean_pinball_loss
except ImportError:
    pass



# (block) bootstrap
def bootstrap(x, h, block_size=None, seed=123):
    """
    Generates block bootstrap indices for a given time series.

    Parameters:
    - x: numpy array, the original time series (univariate or multivariate).
    - h: int, output length
    - block_size: int, the size of the blocks to resample (if None, independent bootstrap).
    - seed: int, reproducibility seed.

    Returns:
    - numpy arrays containing resampled time series.
    """
    if len(x.shape) == 1:
        time_series_length = len(x)
        ndim = 1
    else:
        time_series_length = x.shape[0]
        ndim = x.shape[1]

    if block_size is not None:

        num_blocks = (time_series_length + block_size - 1) // block_size
        all_indices = np.arange(time_series_length)

        indices = []
        for i in range(num_blocks):
            np.random.seed(seed + i * 100)
            start_index = np.random.randint(
                0, time_series_length - block_size + 1
            )
            block_indices = all_indices[start_index: start_index + block_size]
            indices.extend(block_indices)

    else:  # block_size is None

        indices = np.random.choice(
            range(time_series_length), size=h, replace=True
        )

    if ndim == 1:
        return x[np.array(indices[:h])]
    else:
        return x[np.array(indices[:h]), :]


# compute input dates from data frame's index
def compute_input_dates(df):
    input_dates = df.index.values

    frequency = pd.infer_freq(pd.DatetimeIndex(input_dates))

    input_dates = pd.date_range(
        start=input_dates[0], periods=len(input_dates), freq=frequency
    ).values.tolist()

    df_input_dates = pd.DataFrame({"date": input_dates})

    return pd.to_datetime(df_input_dates["date"]).dt.date


# compute output dates from data frame's index
def compute_output_dates(df, horizon):
    input_dates = df.index.values

    if input_dates[0] == 0:
        input_dates = pd.date_range(
            start=pd.Timestamp.today().strftime("%Y-%m-%d"), periods=horizon
        )

    # print(f"\n in nnetsauce.utils.timeseries 1: {input_dates} \n")

    frequency = pd.infer_freq(pd.DatetimeIndex(input_dates))

    # print(f"\n in nnetsauce.utils.timeseries 2: {frequency} \n")

    output_dates = np.delete(
        pd.date_range(
            start=input_dates[-1], periods=horizon + 1, freq=frequency
        ).values,
        0,
    ).tolist()

    # print(f"\n in nnetsauce.utils.timeseries 3: {output_dates} \n")

    df_output_dates = pd.DataFrame({"date": output_dates})
    output_dates = pd.to_datetime(df_output_dates["date"]).dt.date

    return output_dates, frequency


def coverage(obj, actual, level=95, per_series=False):

    alpha = 1 - level / 100
    lt = obj.lower
    ut = obj.upper
    n_points = actual.shape[0]

    if isinstance(lt, pd.DataFrame):
        actual = actual.values

    # Ensure the arrays have the same length
    assert (
        n_points == lt.shape[0] == ut.shape[0]
    ), "actual, lower and upper bounds have different shapes"

    if isinstance(lt, pd.DataFrame) and isinstance(ut, pd.DataFrame):
        diff_lt = (lt.values <= actual) * (ut.values >= actual)
    else:
        diff_lt = (lt <= actual) * (ut >= actual)

    if per_series:
        return np.mean(diff_lt, axis=0) * 100
    else:
        return np.mean(diff_lt) * 100


# create lags for one series
def create_lags(x, k, n=None):
    k_ = k + 1

    n_k = len(x) - k_

    x_ = x[::-1]

    z = [x_[i: (n_k + i + 1)] for i in range(k_)]

    if n is None:
        return np.column_stack(z)

    temp = np.column_stack(z)
    return np.repeat(temp, n).reshape(temp.shape[0], -1)


# create inputs for training from MTS (envisage other regressors in X)
# X in decreasing order (!)
# a = np.reshape(range(0, 24), (8, 3))
# create_train_inputs(a, 2)
def create_train_inputs(X, k):
    n_k = X.shape[0] - k

    z = [X[i: n_k + i, :] for i in range(1, (k + 1))]

    return (X[0:n_k, :], np.column_stack(z))


# reformat response in prediction loop
# a = np.reshape(range(0, 24), (8, 3))
# reformat_response(a, 2)
def reformat_response(X, k):
    return np.concatenate([X[i, :] for i in range(k)])


def winkler_score(obj, actual, level=95, per_series=False):

    alpha = 1 - level / 100
    lt = obj.lower
    ut = obj.upper
    n_points = actual.shape[0]

    # Ensure the arrays have the same length
    assert (
        n_points == lt.shape[0] == ut.shape[0]
    ), "actual, lower and upper bounds have different shapes"

    if isinstance(lt, pd.DataFrame) and isinstance(actual, pd.DataFrame):
        diff_lt = lt.values - actual.values
    else:
        diff_lt = lt - actual
    if isinstance(lt, pd.DataFrame) and isinstance(ut, pd.DataFrame):
        diff_bounds = ut.values - lt.values
    else:
        diff_bounds = ut - lt
    if isinstance(lt, pd.DataFrame) and isinstance(ut, pd.DataFrame):
        diff_ut = actual.values - ut.values
    else:
        diff_ut = actual - ut

    score = diff_bounds + (2 / alpha) * (
        np.maximum(diff_lt, 0) + np.maximum(diff_ut, 0)
    )

    if per_series == True:
        return np.mean(score, axis=0)
    else:
        return np.mean(score)


def mean_errors(
    pred, actual, scoring="root_mean_squared_error", per_series=False
):

    if isinstance(pred, pd.DataFrame):
        pred = pred.values
    else:
        pred = pred.mean.values

    if isinstance(actual, pd.DataFrame):
        actual = actual.values

    diff = pred - actual

    if per_series == True:

        if scoring == "mean_error":
            return np.mean(diff, axis=0).tolist()
        elif scoring == "mean_absolute_error":
            return np.mean(np.abs(diff), axis=0).tolist()
        elif scoring == "mean_percentage_error":
            return np.asarray(np.mean(diff / actual, axis=0) * 100).tolist()
        elif scoring == "mean_absolute_percentage_error":
            return np.asarray(
                np.mean(np.abs(diff / actual), axis=0) * 100
            ).tolist()
        elif scoring == "root_mean_squared_error":
            return np.sqrt(np.mean(np.square(diff), axis=0)).tolist()
        elif scoring == "mean_squared_error":
            return np.mean(np.square(diff), axis=0).tolist()
        elif scoring == "mean_pinball_loss":
            return [
                mean_pinball_loss(actual[:, i], pred[:, i])
                for i in range(actual.shape[1])
            ]

    else:

        if scoring == "mean_error":
            return np.mean(diff)
        elif scoring == "mean_absolute_error":
            return np.mean(np.abs(diff))
        elif scoring == "mean_percentage_error":
            return np.mean(diff / actual) * 100
        elif scoring == "mean_absolute_percentage_error":
            return np.mean(np.abs(diff / actual)) * 100
        elif scoring == "root_mean_squared_error":
            return np.sqrt(np.mean(np.square(diff)))
        elif scoring == "mean_squared_error":
            return np.mean(np.square(diff))
        elif scoring == "mean_pinball_loss":
            return mean_pinball_loss(actual, pred)


# create k lags for series x
# def create_lags(x, k):
#
#    n = len(x)
#    n_k = n - k
#    n_col = k + 1
#    res = np.zeros((n_k, n_col))
#
#    for i in range(n_k):
#        for j in range(n_col):
#            res[i, j] = x[i + j]
#
#    return res


# create inputs for training from MTS (envisage other regressors in X)
# X in decreasing order (!)
# def create_train_inputs(X, k):
#
#    n, p = X.shape
#    n_k = n - k
#
#    y = np.zeros((n_k, p))
#    xreg = np.zeros((n_k, k*p))
#
#    for j in range(p):
#
#        Y_x = create_lags(X[:,j], k)
#
#        y[:,j] = Y_x[:, 0]
#
#        for l in range(1, k):
#
#            xreg[:, j*k + (l-1)] = Y_x[:, l]
#
#        xreg[:, j*k + (k-1)] = Y_x[:, k]
#
#    return (y, xreg)


# reformat response in prediction loop
# def reformat_response(X, k):
#
#    n, p = X.shape
#    res = np.zeros((1, k*p))
#
#    for j in range(p):
#        for i in range(k):
#            res[0, j*k + i] = X[i, j]
#
#    return res
#



import numpy as np
import importlib
import subprocess
import sys


def install_package(package_name):
    """Install a package dynamically using pip."""
    subprocess.check_call([sys.executable, "-m", "pip", "install", package_name])

def check_and_install(package_name):
    """Check if a package is installed; if not, install it."""
    try:
        # Check if the package is already installed by importing it
        importlib.import_module(package_name)        
    except ImportError:
        print(f"'{package_name}' not found. Installing...")
        install_package(package_name)
        # Retry importing the package after installation
        importlib.import_module(package_name)
        print(f"'{package_name}' has been installed successfully.")

def vinecopula_sample(x, n_samples=10, method="vine-tll", random_state=123):
    check_and_install(pyvinecopulib)
    u = pyvinecopulib.to_pseudo_obs(x)
    method_name = method.replace("scp-vine-", "")
    method_name = method_name.replace("scp2-vine-", "")
    method_name = method_name.replace("vine-", "")    
    controls = pyvinecopulib.FitControlsVinecop(
        family_set=[getattr(pyvinecopulib.BicopFamily, method_name)]
    )
    cop = pyvinecopulib.Vinecop(u, controls=controls)
    u_sim = cop.simulate(n_samples, seeds=[random_state])
    p = x.shape[1]
    return np.asarray([np.quantile(x[:, i], u_sim[:, i]) for i in range(p)]).T



import numpy as np
import pandas as pd
try:
    import xgboost as xgb
except ImportError:
    pass
from sklearn.ensemble import RandomForestRegressor
from copy import deepcopy
from tqdm import tqdm
import time
try: 
    from sklearn.utils import all_estimators
except ImportError:
    pass
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import StandardScaler, OneHotEncoder, OrdinalEncoder
from sklearn.compose import ColumnTransformer
from sklearn.base import RegressorMixin
from sklearn.metrics import r2_score
from .config import DEEPREGRESSORS
from ..custom import Custom, CustomRegressor

import warnings

warnings.filterwarnings("ignore")
pd.set_option("display.precision", 2)
pd.set_option("display.float_format", lambda x: "%.2f" % x)

numeric_transformer = Pipeline(
    steps=[
        ("imputer", SimpleImputer(strategy="mean")),
        ("scaler", StandardScaler()),
    ]
)

try:
    categorical_transformer_low = Pipeline(
        steps=[
            (
                "imputer",
                SimpleImputer(strategy="constant", fill_value="missing"),
            ),
            (
                "encoding",
                OneHotEncoder(handle_unknown="ignore", sparse_output=False),
            ),
        ]
    )
except TypeError:
    categorical_transformer_low = Pipeline(
        steps=[
            (
                "imputer",
                SimpleImputer(strategy="constant", fill_value="missing"),
            ),
            (
                "encoding",
                OneHotEncoder(handle_unknown="ignore", sparse=False),
            ),
        ]
    )

categorical_transformer_high = Pipeline(
    steps=[
        ("imputer", SimpleImputer(strategy="constant", fill_value="missing")),
        # 'OrdinalEncoder' Raise a ValueError when encounters an unknown value. Check https://github.com/scikit-learn/scikit-learn/pull/13423
        ("encoding", OrdinalEncoder()),
    ]
)


# Helper function


def get_card_split(df, cols, n=11):
    cond = df[cols].nunique() > n
    card_high = cols[cond]
    card_low = cols[~cond]
    return card_low, card_high


def adjusted_rsquared(r2, n, p):
    return 1 - (1 - r2) * ((n - 1) / (n - p - 1))


class LazyDeepRegressor(Custom, RegressorMixin):
    """
        Fitting -- almost -- all the regression algorithms with layers of
        nnetsauce's CustomRegressor and returning their scores.

    Parameters:

        verbose: int, optional (default=0)
            Any positive number for verbosity.

        ignore_warnings: bool, optional (default=True)
            When set to True, the warning related to algorigms that are not able to run are ignored.

        custom_metric: function, optional (default=None)
            When function is provided, models are evaluated based on the custom evaluation metric provided.

        predictions: bool, optional (default=False)
            When set to True, the predictions of all the models models are returned as dataframe.

        sort_by: string, optional (default='RMSE')
            Sort models by a metric. Available options are 'R-Squared', 'Adjusted R-Squared', 'RMSE', 'Time Taken' and 'Custom Metric'.
            or a custom metric identified by its name and provided by custom_metric.

        random_state: int, optional (default=42)
            Reproducibiility seed.

        estimators: list, optional (default='all')
            list of Estimators names or just 'all' (default='all')

        preprocess: bool
            preprocessing is done when set to True

        n_jobs : int, when possible, run in parallel
            For now, only used by individual models that support it.

        n_layers: int, optional (default=3)
            Number of layers of CustomRegressors to be used.

        All the other parameters are the same as CustomRegressor's.
    
    Attributes:

        models_: dict-object
            Returns a dictionary with each model pipeline as value
            with key as name of models.
        
        best_model_: object
            Returns the best model pipeline based on the sort_by metric.                

    Examples:

        import nnetsauce as ns
        import numpy as np
        from sklearn import datasets
        from sklearn.utils import shuffle

        diabetes = datasets.load_diabetes()
        X, y = shuffle(diabetes.data, diabetes.target, random_state=13)
        X = X.astype(np.float32)

        offset = int(X.shape[0] * 0.9)
        X_train, y_train = X[:offset], y[:offset]
        X_test, y_test = X[offset:], y[offset:]

        reg = ns.LazyDeepRegressor(verbose=0, ignore_warnings=False, custom_metric=None)
        models, predictions = reg.fit(X_train, X_test, y_train, y_test)
        print(models)

    """

    def __init__(
        self,
        verbose=0,
        ignore_warnings=True,
        custom_metric=None,
        predictions=False,
        sort_by="RMSE",
        random_state=42,
        estimators="all",
        preprocess=False,
        n_jobs=None,
        # Defining depth
        n_layers=3,
        # CustomRegressor attributes
        obj=None,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
        backend="cpu",
    ):
        self.verbose = verbose
        self.ignore_warnings = ignore_warnings
        self.custom_metric = custom_metric
        self.predictions = predictions
        self.sort_by = sort_by
        self.models_ = {}
        self.best_model_ = None
        self.random_state = random_state
        self.estimators = estimators
        self.preprocess = preprocess
        self.n_layers = n_layers - 1
        self.n_jobs = n_jobs
        super().__init__(
            obj=obj,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
            backend=backend,
        )

    def fit(self, X_train, X_test, y_train, y_test):
        """Fit Regression algorithms to X_train and y_train, predict and score on X_test, y_test.

        Parameters:

            X_train : array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            X_test : array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

            y_train : array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            y_test : array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

        Returns:
        -------
        scores:  Pandas DataFrame
            Returns metrics of all the models in a Pandas DataFrame.

        predictions : Pandas DataFrame
            Returns predictions of all the models in a Pandas DataFrame.

        """
        R2 = []
        ADJR2 = []
        RMSE = []
        # WIN = []
        names = []
        TIME = []
        predictions = {}

        if self.custom_metric:
            CUSTOM_METRIC = []

        if isinstance(X_train, np.ndarray):
            X_train = pd.DataFrame(X_train)
            X_test = pd.DataFrame(X_test)

        numeric_features = X_train.select_dtypes(include=[np.number]).columns
        categorical_features = X_train.select_dtypes(include=["object"]).columns

        categorical_low, categorical_high = get_card_split(
            X_train, categorical_features
        )

        if self.preprocess is True:
            preprocessor = ColumnTransformer(
                transformers=[
                    ("numeric", numeric_transformer, numeric_features),
                    (
                        "categorical_low",
                        categorical_transformer_low,
                        categorical_low,
                    ),
                    (
                        "categorical_high",
                        categorical_transformer_high,
                        categorical_high,
                    ),
                ]
            )
        
        # base models
        try: 
            baseline_names = ["RandomForestRegressor", "XGBRegressor"]
            baseline_models = [RandomForestRegressor(), xgb.XGBRegressor()]
        except Exception as exception:
            baseline_names = ["RandomForestRegressor"]
            baseline_models = [RandomForestRegressor()]

        for name, model in zip(baseline_names, baseline_models):        
            start = time.time()
            try:
                model.fit(X_train, y_train)
                self.models_[name] = model
                y_pred = model.predict(X_test)
                r_squared = r2_score(y_test, y_pred)
                adj_rsquared = adjusted_rsquared(
                    r_squared, X_test.shape[0], X_test.shape[1]
                )
                rmse = np.sqrt(np.mean((y_test - y_pred)**2))

                names.append(name)
                R2.append(r_squared)
                ADJR2.append(adj_rsquared)
                RMSE.append(rmse)
                TIME.append(time.time() - start)

                if self.custom_metric:
                    custom_metric = self.custom_metric(y_test, y_pred)
                    CUSTOM_METRIC.append(custom_metric)

                if self.verbose > 0:
                    scores_verbose = {
                        "Model": name,
                        "R-Squared": r_squared,
                        "Adjusted R-Squared": adj_rsquared,
                        "RMSE": rmse,
                        "Time taken": time.time() - start,
                    }

                    if self.custom_metric:
                        scores_verbose[self.custom_metric.__name__] = (
                            custom_metric
                        )

                    print(scores_verbose)
                if self.predictions:
                    predictions[name] = y_pred
            except Exception as exception:
                if self.ignore_warnings is False:
                    print(name + " model failed to execute")
                    print(exception)                

        if self.estimators == "all":
            self.regressors = DEEPREGRESSORS
        else:
            self.regressors = [
                ("DeepCustomRegressor(" + est[0] + ")", est[1])
                for est in all_estimators()
                if (
                    issubclass(est[1], RegressorMixin)
                    and (est[0] in self.estimators)
                )
            ]

        if self.preprocess is True:

            for name, model in tqdm(self.regressors):  # do parallel exec
                start = time.time()
                try:
                    if "random_state" in model().get_params().keys():
                        layer_regr = CustomRegressor(
                            obj=model(random_state=self.random_state),
                            n_hidden_features=self.n_hidden_features,
                            activation_name=self.activation_name,
                            a=self.a,
                            nodes_sim=self.nodes_sim,
                            bias=self.bias,
                            dropout=self.dropout,
                            direct_link=self.direct_link,
                            n_clusters=self.n_clusters,
                            cluster_encode=self.cluster_encode,
                            type_clust=self.type_clust,
                            type_scaling=self.type_scaling,
                            col_sample=self.col_sample,
                            row_sample=self.row_sample,
                            seed=self.seed,
                            backend=self.backend,
                        )
                    else:
                        layer_regr = CustomRegressor(
                            obj=model(),
                            n_hidden_features=self.n_hidden_features,
                            activation_name=self.activation_name,
                            a=self.a,
                            nodes_sim=self.nodes_sim,
                            bias=self.bias,
                            dropout=self.dropout,
                            direct_link=self.direct_link,
                            n_clusters=self.n_clusters,
                            cluster_encode=self.cluster_encode,
                            type_clust=self.type_clust,
                            type_scaling=self.type_scaling,
                            col_sample=self.col_sample,
                            row_sample=self.row_sample,
                            seed=self.seed,
                            backend=self.backend,
                        )

                    for _ in range(self.n_layers):
                        layer_regr = deepcopy(
                            CustomRegressor(
                                obj=layer_regr,
                                n_hidden_features=self.n_hidden_features,
                                activation_name=self.activation_name,
                                a=self.a,
                                nodes_sim=self.nodes_sim,
                                bias=self.bias,
                                dropout=self.dropout,
                                direct_link=self.direct_link,
                                n_clusters=self.n_clusters,
                                cluster_encode=self.cluster_encode,
                                type_clust=self.type_clust,
                                type_scaling=self.type_scaling,
                                col_sample=self.col_sample,
                                row_sample=self.row_sample,
                                seed=self.seed,
                                backend=self.backend,
                            )
                        )

                    layer_regr.fit(X_train, y_train)

                    pipe = Pipeline(
                        steps=[
                            ("preprocessor", preprocessor),
                            ("regressor", layer_regr),
                        ]
                    )

                    pipe.fit(X_train, y_train)

                    self.models_[name] = pipe
                    y_pred = pipe.predict(X_test)
                    r_squared = r2_score(y_test, y_pred)
                    adj_rsquared = adjusted_rsquared(
                        r_squared, X_test.shape[0], X_test.shape[1]
                    )
                    rmse = np.sqrt(np.mean((y_test - y_pred)**2))

                    names.append(name)
                    R2.append(r_squared)
                    ADJR2.append(adj_rsquared)
                    RMSE.append(rmse)
                    TIME.append(time.time() - start)

                    if self.custom_metric:
                        custom_metric = self.custom_metric(y_test, y_pred)
                        CUSTOM_METRIC.append(custom_metric)

                    if self.verbose > 0:
                        scores_verbose = {
                            "Model": name,
                            "R-Squared": r_squared,
                            "Adjusted R-Squared": adj_rsquared,
                            "RMSE": rmse,
                            "Time taken": time.time() - start,
                        }

                        if self.custom_metric:
                            scores_verbose[self.custom_metric.__name__] = (
                                custom_metric
                            )

                        print(scores_verbose)
                    if self.predictions:
                        predictions[name] = y_pred
                except Exception as exception:
                    if self.ignore_warnings is False:
                        print(name + " model failed to execute")
                        print(exception)

        else:  # no preprocessing

            for name, model in tqdm(self.regressors):  # do parallel exec
                start = time.time()
                try:
                    if "random_state" in model().get_params().keys():
                        layer_regr = CustomRegressor(
                            obj=model(random_state=self.random_state),
                            n_hidden_features=self.n_hidden_features,
                            activation_name=self.activation_name,
                            a=self.a,
                            nodes_sim=self.nodes_sim,
                            bias=self.bias,
                            dropout=self.dropout,
                            direct_link=self.direct_link,
                            n_clusters=self.n_clusters,
                            cluster_encode=self.cluster_encode,
                            type_clust=self.type_clust,
                            type_scaling=self.type_scaling,
                            col_sample=self.col_sample,
                            row_sample=self.row_sample,
                            seed=self.seed,
                            backend=self.backend,
                        )
                    else:
                        layer_regr = CustomRegressor(
                            obj=model(),
                            n_hidden_features=self.n_hidden_features,
                            activation_name=self.activation_name,
                            a=self.a,
                            nodes_sim=self.nodes_sim,
                            bias=self.bias,
                            dropout=self.dropout,
                            direct_link=self.direct_link,
                            n_clusters=self.n_clusters,
                            cluster_encode=self.cluster_encode,
                            type_clust=self.type_clust,
                            type_scaling=self.type_scaling,
                            col_sample=self.col_sample,
                            row_sample=self.row_sample,
                            seed=self.seed,
                            backend=self.backend,
                        )

                    layer_regr.fit(X_train, y_train)

                    for _ in range(self.n_layers):
                        layer_regr = deepcopy(
                            CustomRegressor(
                                obj=layer_regr,
                                n_hidden_features=self.n_hidden_features,
                                activation_name=self.activation_name,
                                a=self.a,
                                nodes_sim=self.nodes_sim,
                                bias=self.bias,
                                dropout=self.dropout,
                                direct_link=self.direct_link,
                                n_clusters=self.n_clusters,
                                cluster_encode=self.cluster_encode,
                                type_clust=self.type_clust,
                                type_scaling=self.type_scaling,
                                col_sample=self.col_sample,
                                row_sample=self.row_sample,
                                seed=self.seed,
                                backend=self.backend,
                            )
                        )

                        # layer_regr.fit(X_train, y_train)

                    layer_regr.fit(X_train, y_train)

                    self.models_[name] = layer_regr
                    y_pred = layer_regr.predict(X_test)

                    r_squared = r2_score(y_test, y_pred)
                    adj_rsquared = adjusted_rsquared(
                        r_squared, X_test.shape[0], X_test.shape[1]
                    )
                    rmse = np.sqrt(np.mean((y_test - y_pred)**2))

                    names.append(name)
                    R2.append(r_squared)
                    ADJR2.append(adj_rsquared)
                    RMSE.append(rmse)
                    TIME.append(time.time() - start)

                    if self.custom_metric:
                        custom_metric = self.custom_metric(y_test, y_pred)
                        CUSTOM_METRIC.append(custom_metric)

                    if self.verbose > 0:
                        scores_verbose = {
                            "Model": name,
                            "R-Squared": r_squared,
                            "Adjusted R-Squared": adj_rsquared,
                            "RMSE": rmse,
                            "Time taken": time.time() - start,
                        }

                        if self.custom_metric:
                            scores_verbose[self.custom_metric.__name__] = (
                                custom_metric
                            )

                        print(scores_verbose)
                    if self.predictions:
                        predictions[name] = y_pred
                except Exception as exception:
                    if self.ignore_warnings is False:
                        print(name + " model failed to execute")
                        print(exception)

        scores = {
            "Model": names,
            "Adjusted R-Squared": ADJR2,
            "R-Squared": R2,
            "RMSE": RMSE,
            "Time Taken": TIME,
        }

        if self.custom_metric:
            scores["Custom metric"] = CUSTOM_METRIC

        scores = pd.DataFrame(scores)
        scores = scores.sort_values(by=self.sort_by, ascending=True).set_index(
            "Model"
        )

        self.best_model_ = self.models_[scores.index[0]]

        if self.predictions is True:
            
            return scores, predictions
        
        return scores

    def get_best_model(self):
        """
        This function returns the best model pipeline based on the sort_by metric.

        Returns:

            best_model: object,
                Returns the best model pipeline based on the sort_by metric.

        """
        return self.best_model_
    
    def provide_models(self, X_train, X_test, y_train, y_test):
        """
        This function returns all the model objects trained in fit function.
        If fit is not called already, then we call fit and then return the models.

        Parameters:

            X_train : array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            X_test : array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

            y_train : array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            y_test : array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

        Returns:

            models: dict-object,
                Returns a dictionary with each model pipeline as value
                with key as name of models.

        """
        if len(self.models_.keys()) == 0:
            self.fit(X_train, X_test, y_train, y_test)

        return self.models_


class LazyRegressor(LazyDeepRegressor):
    """
        Fitting -- almost -- all the regression algorithms with 
        nnetsauce's CustomRegressor and returning their scores.

    Parameters:

        verbose: int, optional (default=0)
            Any positive number for verbosity.

        ignore_warnings: bool, optional (default=True)
            When set to True, the warning related to algorigms that are not able to run are ignored.

        custom_metric: function, optional (default=None)
            When function is provided, models are evaluated based on the custom evaluation metric provided.

        predictions: bool, optional (default=False)
            When set to True, the predictions of all the models models are returned as dataframe.

        sort_by: string, optional (default='RMSE')
            Sort models by a metric. Available options are 'R-Squared', 'Adjusted R-Squared', 'RMSE', 'Time Taken' and 'Custom Metric'.
            or a custom metric identified by its name and provided by custom_metric.

        random_state: int, optional (default=42)
            Reproducibiility seed.

        estimators: list, optional (default='all')
            list of Estimators names or just 'all' (default='all')

        preprocess: bool
            preprocessing is done when set to True

        n_jobs : int, when possible, run in parallel
            For now, only used by individual models that support it.

        All the other parameters are the same as CustomRegressor's.
    
    Attributes:

        models_: dict-object
            Returns a dictionary with each model pipeline as value
            with key as name of models.
        
        best_model_: object
            Returns the best model pipeline based on the sort_by metric.                

    Examples:

        import nnetsauce as ns
        import numpy as np
        from sklearn import datasets
        from sklearn.utils import shuffle

        diabetes = datasets.load_diabetes()
        X, y = shuffle(diabetes.data, diabetes.target, random_state=13)
        X = X.astype(np.float32)

        offset = int(X.shape[0] * 0.9)
        X_train, y_train = X[:offset], y[:offset]
        X_test, y_test = X[offset:], y[offset:]

        reg = ns.LazyRegressor(verbose=0, ignore_warnings=False,
                            custom_metric=None)
        models, predictions = reg.fit(X_train, X_test, y_train, y_test)
        print(models)
    
    """
    
    def __init__(
        self,
        verbose=0,
        ignore_warnings=True,
        custom_metric=None,
        predictions=False,
        sort_by="RMSE",
        random_state=42,
        estimators="all",
        preprocess=False,
        n_jobs=None,
        # CustomRegressor attributes
        obj=None,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
        backend="cpu",
    ):
        super().__init__(
            verbose=verbose,
            ignore_warnings=ignore_warnings,
            custom_metric=custom_metric,
            predictions=predictions,
            sort_by=sort_by,
            random_state=random_state,
            estimators=estimators,
            preprocess=preprocess,
            n_jobs=n_jobs,
            n_layers=1,
            obj=obj,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
            backend=backend,
        )


# Authors: Thierry Moudiki
#
# License: BSD 3 Clear

import numpy as np
from joblib import Parallel, delayed
from tqdm import tqdm
from ..utils.misc import flatten, is_factor


def dosubsample(y, row_sample=0.8, seed=123, n_jobs=None, verbose=False):
    index = []
    assert (row_sample < 1) & (
        row_sample >= 0
    ), "'row_sample' must be < 1 and >= 0"
    n_obs = len(y)
    n_obs_out = np.ceil(n_obs * row_sample)

    # preproc -----
    if is_factor(y):  # classification

        classes, n_elem_classes = np.unique(y, return_counts=True)
        n_classes = len(classes)
        y_as_classes = y.copy()
        freqs_hist = np.zeros_like(n_elem_classes, dtype=float)
        if verbose is True:
            print(f"creating breaks...")
        if n_jobs is None:
            for i in range(len(n_elem_classes)):
                freqs_hist[i] = float(n_elem_classes[i]) / n_obs
        else:

            def get_freqs_hist(i):
                return float(n_elem_classes[i]) / n_obs

            freqs_hist = Parallel(n_jobs=n_jobs)(
                delayed(get_freqs_hist)(i) for i in range(len(n_elem_classes))
            )

    else:  # regression

        h = np.histogram(y, bins="auto")
        n_elem_classes = np.asarray(h[0], dtype=np.int32)
        freqs_hist = np.zeros_like(n_elem_classes, dtype=float)

        if verbose is True:
            print(f"creating breaks...")
        if n_jobs is None:
            for i in range(len(n_elem_classes)):
                freqs_hist[i] = float(n_elem_classes[i]) / n_obs
        else:

            def get_freqs_hist(i):
                return float(n_elem_classes[i]) / n_obs

            freqs_hist = Parallel(n_jobs=n_jobs)(
                delayed(get_freqs_hist)(i) for i in range(len(n_elem_classes))
            )

        breaks = h[1]
        n_breaks_1 = len(breaks) - 1
        classes = range(n_breaks_1)
        n_classes = n_breaks_1
        y_as_classes = np.zeros_like(y, dtype=int)

        for i in classes:
            y_as_classes[(y > breaks[i]) * (y <= breaks[i + 1])] = int(i)

    # main loop ----

    if verbose is True:
        print(f"main loop...")

    if n_jobs is None:

        if verbose is True:
            iterator = tqdm(range(n_classes))
        else:
            iterator = range(n_classes)

        for i in iterator:
            bool_class_i = y_as_classes == classes[i]
            index_class_i = np.asarray(
                np.where(bool_class_i == True)[0], dtype=np.int32
            )
            if np.sum(bool_class_i) > 1:  # at least 2 elements in class  #i
                np.random.seed(seed + i)
                index.extend(
                    np.random.choice(
                        index_class_i,
                        size=int(n_obs_out * freqs_hist[i]),  # output size
                        replace=True,
                    ).tolist()
                )
            else:  # only one element in class
                try:
                    index.append(index_class_i[0])
                except:
                    pass

    else:  # parallel execution

        def get_index(i):
            bool_class_i = y_as_classes == classes[i]
            index_class_i = np.asarray(
                np.where(bool_class_i == True)[0], dtype=np.int32
            )
            if np.sum(bool_class_i) > 1:  # at least 2 elements in class  #i
                np.random.seed(seed + i)
                return np.random.choice(
                    index_class_i,
                    size=int(n_obs_out * freqs_hist[i]),  # output size
                    replace=True,
                ).tolist()
            else:  # only one element in class
                try:
                    return index_class_i[0]
                except:
                    pass

        if verbose is True:
            index = Parallel(n_jobs=n_jobs)(
                delayed(get_index)(i) for i in tqdm(range(n_classes))
            )
        else:
            index = Parallel(n_jobs=n_jobs)(
                delayed(get_index)(i) for i in range(n_classes)
            )

    try:
        return np.asarray(flatten(index))
    except:
        return np.asarray(index)
